<?php
/**
 * @version        6.2.0
 * @package        Joomla
 * @subpackage     Payment Form
 * @author         Tuan Pham Ngoc
 * @copyright      Copyright (C) 2010 - 2021 Ossolution Team
 * @license        GNU/GPL, see LICENSE.php
 */

namespace OSSolution\PaymentForm\Site\Controller;

use Joomla\CMS\Captcha\Captcha;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\HTML\HTMLHelper;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Mail\MailHelper;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\CMS\Uri\Uri;
use OSL\Controller\Download;
use OSSolution\PaymentForm\Site\Helper\Helper as PMFormHelper;
use OSSolution\PaymentForm\Site\Helper\PaymentMethods as PaymentMethodsHelper;

defined('_JEXEC') or die;

class Controller extends \OSL\Controller\Controller
{
	use Download;

	/**
	 * Display information
	 *
	 */
	public function display($cachable = false, array $urlparams = [])
	{
		$document = $this->container->document;
		$config   = PMFormHelper::getConfig();
		$rootUri  = Uri::root(true);
		$document->addStyleSheet($rootUri . '/media/com_pmform/assets/css/style.min.css', 'text/css', null, null);

		if (file_exists(JPATH_ROOT . '/media/com_pmform/assets/css/custom.css') && filesize(JPATH_ROOT . '/media/com_pmform/assets/css/custom.css') > 0)
		{
			$document->addStyleSheet($rootUri . '/media/com_pmform/assets/css/custom.css');
		}

		HTMLHelper::_('jquery.framework');

		if ($config->load_twitter_bootstrap !== '0')
		{
			PMFormHelper::loadBootstrap(false);
		}

		$document->addScript($rootUri . '/media/com_pmform/assets/js/noconflict.min.js');

		//Clear the form data on payment complete or cancel
		$viewName = $this->input->get('view', $this->container->defaultView, 'string');

		if ($viewName == 'complete' || $viewName == 'cancel')
		{
			$this->container->app->setUserState('com_pmform.formdata', null);
		}

		parent::display($cachable, $urlparams);
	}

	/**
	 * Process Payment
	 *
	 */
	public function process_payment()
	{
		$this->csrfProtection();

		/* @var \JApplicationSite $app */
		$app     = $this->container->get('app');
		$session = $this->container->session;
		$user    = $this->container->user;
		$config  = PMFormHelper::getConfig();

		$app->setUserState('com_pmform.formdata', serialize($this->input->getData()));

		$errors = $this->validateFormData($this->input->getInt('form_id'), $this->input->getData());

		// Validate username and password
		if (!$user->id && $config->user_registration)
		{
			$errors = array_merge($errors, $this->validateUsername($this->input->post->get('username', '', 'raw')));
			$errors = array_merge($errors, $this->validatePassword($this->input->post->get('password1', '', 'raw')));
			$errors = array_merge($errors, $this->validateEmail($this->input->post->get('email', '', 'raw')));
		}

		// Check captcha if captcha is enabled
		$captchaErrorMessage = '';

		if ($config->enable_captcha)
		{
			if ($config->show_confirmation_step)
			{
				$res = $session->get('pf_captcha_valid', 0);
			}
			else
			{
				$captchaPlugin = $app->getParams()->get('captcha', $this->container->appConfig->get('captcha'));

				if (!$captchaPlugin)
				{
					// Hardcode to recaptcha, reduce support request
					$captchaPlugin = 'recaptcha';
				}

				$plugin = PluginHelper::getPlugin('captcha', $captchaPlugin);

				if ($plugin)
				{
					try
					{
						$res = Captcha::getInstance($captchaPlugin)->checkAnswer($this->input->post->get('recaptcha_response_field', '', 'string'));
					}
					catch (\Exception $e)
					{
						$captchaErrorMessage = $e->getMessage();
						$res                 = false;
					}
				}
				else
				{
					$res = true;
				}
			}

			if (!$res)
			{
				$errors[] = $captchaErrorMessage;
				$session->set('pf_captcha_invalid', 1);
			}
		}

		$formPageUrl = $this->input->get('form_page_url', '', 'none');

		if ($formPageUrl)
		{
			$session->set('form_page_url', $formPageUrl);
		}

		if (count($errors))
		{
			foreach ($errors as $error)
			{
				$app->enqueueMessage($error, 'error');
			}

			// Redirect back to the form
			$app->redirect(base64_decode($formPageUrl));

			return false;
		}
		else
		{
			$session->clear('pf_captcha_invalid');
			$session->clear('pf_captcha_valid');
		}

		/* @var \OSSolution\PaymentForm\Site\Model\Payment $model */
		$model = $this->getModel('Payment');

		$model->processPayment($this->input);
	}

	/**
	 * Display payment confirmation page
	 *
	 * @return bool
	 */
	public function payment_confirmation()
	{
		$this->csrfProtection();

		/* @var \JApplicationSite $app */
		$app    = $this->container->get('app');
		$config = PMFormHelper::getConfig();

		$app->setUserState('com_pmform.formdata', serialize($this->input->getData()));

		// Check captcha if captcha is enabled
		if ($config->enable_captcha)
		{
			$session = $this->container->session;

			$captchaPlugin = $app->getParams()->get('captcha', $this->container->appConfig->get('captcha'));

			if (!$captchaPlugin)
			{
				// Hardcode to recaptcha, reduce support request
				$captchaPlugin = 'recaptcha';
			}

			$plugin = PluginHelper::getPlugin('captcha', $captchaPlugin);

			if ($plugin)
			{
				$res = Captcha::getInstance($captchaPlugin)->checkAnswer($this->input->post->get('recaptcha_response_field', '', 'string'));
			}
			else
			{
				$res = true;
			}

			if (!$res)
			{
				$session->set('pf_captcha_invalid', 1);
				$app->enqueueMessage(Text::_('PF_INVALID_CAPTCHA_ENTERED'), 'error');
				$formPageUrl = $this->input->get('form_page_url', '', 'none');
				$app->redirect(base64_decode($formPageUrl));

				return false;
			}
			else
			{
				$session->clear('pf_captcha_invalid');
				$session->set('pf_captcha_valid', 1);
			}
		}

		// Captcha is valid or not activated, we will display the payment confirmation page
		$this->input->set('view', 'confirmation');
		$this->input->set('layout', 'default');
		$this->display();
	}

	/**
	 * Payment Confirmation
	 *
	 */
	public function payment_confirm()
	{
		$paymentMethod = $this->input->get('payment_method', '', 'none');
		$method        = PaymentMethodsHelper::getPaymentMethod($paymentMethod);

		if ($method)
		{
			$method->verifyPayment();
		}
	}

	/**
	 * Process downloading a file
	 *
	 */
	public function download_file()
	{
		$uploadPath = PMFormHelper::getConfigValue('path_upload') ?: 'images/com_pmform';
		$fileName   = $this->input->get('file_name', '', 'none');
		$filePath   = JPATH_ROOT . '/' . $uploadPath . '/' . $fileName;

		if ($fileName && file_exists($filePath))
		{
			$fileName = PMFormHelper::getOriginalFilename($fileName);

			$this->processDownloadFile($filePath, $fileName);
		}
		else
		{
			$this->container->app->enqueueMessage(Text::_('PF_FILE_NOT_EXIST'), 'error');
			$this->container->app->redirect('index.php?option=com_pmform&Itemid=' . $this->input->getInt('Itemid', 0), 404);
		}
	}

	/**
	 * Get depend fields status to show/hide custom fields based on selected options
	 */
	public function get_depend_fields_status()
	{
		$input   = $this->input;
		$db      = $this->container->db;
		$fieldId = $this->input->get('field_id', 'int');

		$masterFields = [];
		$fieldsAssoc  = [];
		$hiddenFields = [];
		$showFields   = [];
		$hideFields   = [];

		//Get list of depend fields
		$allFieldIds = PMFormHelper::getAllDependencyFields($fieldId);
		$query       = $db->getQuery(true);
		$query->select('*')
			->from('#__pf_fields')
			->where('id IN (' . implode(',', $allFieldIds) . ')')
			->where('published = 1')
			->order('ordering');

		$db->setQuery($query);
		$rowFields = $db->loadObjectList();

		foreach ($rowFields as $rowField)
		{
			if ($rowField->depend_on_field_id > 0)
			{
				$masterFields[] = $rowField->depend_on_field_id;
			}

			$fieldsAssoc[$rowField->id] = $rowField;
		}

		$masterFields = array_unique($masterFields);

		if (count($masterFields))
		{
			foreach ($rowFields as $rowField)
			{
				if ($rowField->depend_on_field_id && isset($fieldsAssoc[$rowField->depend_on_field_id]))
				{
					// If master field is hided, then children field should be hided, too
					if (in_array($rowField->depend_on_field_id, $hiddenFields))
					{
						$hiddenFields[] = $rowField->id;
					}
					else
					{
						$fieldName = $fieldsAssoc[$rowField->depend_on_field_id]->name;

						$masterFieldValues = $input->get($fieldName, '', 'none');

						if (is_array($masterFieldValues))
						{
							$selectedOptions = $masterFieldValues;
						}
						else
						{
							$selectedOptions = [$masterFieldValues];
						}

						if (is_string($rowField->depend_on_options) && is_array(json_decode($rowField->depend_on_options)))
						{
							$dependOnOptions = json_decode($rowField->depend_on_options);
						}
						else
						{
							$dependOnOptions = explode(',', $rowField->depend_on_options);
						}

						if (!count(array_intersect($selectedOptions, $dependOnOptions)))
						{
							$hiddenFields[] = $rowField->id;
						}
					}
				}
			}
		}


		foreach ($rowFields as $rowField)
		{
			if (in_array($rowField->id, $hiddenFields))
			{
				$hideFields[] = 'field_' . $rowField->name;
			}
			else
			{
				$showFields[] = 'field_' . $rowField->name;
			}
		}

		echo json_encode(['show_fields' => implode(',', $showFields), 'hide_fields' => implode(',', $hideFields)]);

		$this->container->app->close();
	}

	/**
	 * Calculate registration fee, then update the information on registration form
	 */
	public function calculate_form_fee()
	{
		$config        = PMFormHelper::getConfig();
		$db            = $this->container->db;
		$query         = $db->getQuery(true);
		$input         = $this->input;
		$paymentMethod = $input->getString('payment_method', '');
		$currentDate   = HTMLHelper::_('date', 'now', 'Y-m-d');
		$query->select('*, DATEDIFF("' . $currentDate . '", late_payment_date) AS number_days')
			->from('#__pf_forms')
			->where('id=' . (int) $input->getInt('form_id', 0));
		$db->setQuery($query);
		$rowForm   = $db->loadObject();
		$rowFields = PMFormHelper::getFormFields($input->getInt('form_id', 0));
		$form      = new \POSFForm($rowFields);
		$form->bind($input->getData(), false);
		$fees                               = PMFormHelper::callOverridableHelperMethod('Helper', 'calculateFormFee', [$rowForm, $form, $input, $paymentMethod]);
		$response['total_amount']           = PMFormHelper::formatAmount($fees['total_amount'], $config);
		$response['discount_amount']        = PMFormHelper::formatAmount($fees['discount_amount'], $config);
		$response['tax_amount']             = PMFormHelper::formatAmount($fees['tax_amount'], $config);
		$response['gross_amount']           = PMFormHelper::formatAmount($fees['gross_amount'], $config);
		$response['payment_processing_fee'] = PMFormHelper::formatAmount($fees['payment_processing_fee'], $config);
		$response['coupon_valid']           = $fees['coupon_valid'];

		echo json_encode($response);

		$this->container->app->close();
	}

	/**
	 * Validate username which users entered on order form
	 *
	 */
	public function validate_username()
	{
		$db         = $this->container->db;
		$query      = $db->getQuery(true);
		$username   = $this->input->get('fieldValue', '', 'string');
		$validateId = $this->input->get('fieldId', '', 'string');
		$query->select('COUNT(*)')
			->from('#__users')
			->where('username="' . $username . '"');
		$db->setQuery($query);
		$total        = $db->loadResult();
		$arrayToJs    = [];
		$arrayToJs[0] = $validateId;

		if ($total)
		{
			$arrayToJs[1] = false;
		}
		else
		{
			$arrayToJs[1] = true;
		}

		echo json_encode($arrayToJs);

		$this->container->app->close();
	}

	/**
	 * Validate email which users entered on order form to make sure it is valid
	 */
	public function validate_email()
	{
		$user = $this->container->user;

		$email      = $this->input->get('fieldValue', '', 'string');
		$validateId = $this->input->get('fieldId', '', 'string');
		$formId     = $this->input->getInt('form_id');
		$db         = $this->container->db;
		$query      = $db->getQuery(true)
			->select('*')
			->from('#__pf_forms')
			->where('id = ' . $formId);
		$db->setQuery($query);
		$rowForm = $db->loadObject();

		$arrayToJs    = [];
		$arrayToJs[0] = $validateId;
		$arrayToJs[1] = true;

		$config = PMFormHelper::getConfig();

		if ((int) $rowForm->user_registration !== 2)
		{
			$config->user_registration = $rowForm->user_registration;
		}

		if (!$user->id && $config->user_registration)
		{
			$db    = $this->container->db;
			$query = $db->getQuery(true);
			$query->select('COUNT(*)')
				->from('#__users')
				->where('email = ' . $db->quote($email));
			$db->setQuery($query);
			$total = $db->loadResult();

			if ($total > 0)
			{
				$arrayToJs[1] = false;
			}
		}

		echo json_encode($arrayToJs);

		$this->container->app->close();
	}

	/**
	 * Get list of states for the selected country, using in AJAX request
	 */
	public function get_states()
	{
		$countryName = $this->input->get('country_name', '', 'string');
		$stateName   = $this->input->get('state_name', '', 'string');

		if (!$countryName)
		{
			$countryName = PMFormHelper::getConfigValue('default_country');
		}

		$db    = $this->container->db;
		$query = $db->getQuery(true);
		$query->select('css_class, required')
			->from('#__pf_fields')
			->where('name = ' . $db->quote('state'));
		$db->setQuery($query);
		$rowField = $db->loadObject();

		$classClasses = ['uk-select'];

		if ($rowField->required)
		{
			$classClasses[] = 'validate[required]';
		}

		if ($rowField->css_class)
		{
			$classClasses[] = $rowField->css_class;
		}
		else
		{
			$classClasses[] = 'input-large';
		}

		$class = implode(' ', $classClasses);

		$query->clear()
			->select('country_id')
			->from('#__pf_countries')
			->where('name=' . $db->quote($countryName));
		$db->setQuery($query);
		$countryId = $db->loadResult();

		//get state
		$query->clear()
			->select('state_name AS value, state_name AS text')
			->from('#__pf_states')
			->where('country_id=' . (int) $countryId);;
		$db->setQuery($query);
		$states  = $db->loadObjectList();
		$options = [];

		if (count($states))
		{
			$options[] = HTMLHelper::_('select.option', '', Text::_('PF_SELECT_STATE'));
			$options   = array_merge($options, $states);
		}
		else
		{
			$options[] = HTMLHelper::_('select.option', 'N/A', Text::_('PF_NA'));
		}
		echo HTMLHelper::_('select.genericlist', $options, 'state', ' class="' . $class . ' form-select" id="state" ', 'value', 'text',
			$stateName);

		$this->container->app->close();
	}

	/**
	 * Download invoice of a payment record
	 *
	 * @throws \Exception
	 */
	public function download_invoice()
	{
		$app  = $this->container->app;
		$user = $this->container->user;

		if (!$user->id)
		{
			$app->enqueueMessage(Text::_('You do not have permission to download the invoice'), 'error');
			$app->redirect('index.php', 403);
		}

		$id = $this->input->getInt('id');

		$row = $this->container->factory->createTable('Payment');

		$row->load($id);

		if (!$row->id || ($row->user_id != $user->id))
		{
			$app->enqueueMessage(Text::_('You do not have permission to download the invoice'), 'error');
			$app->redirect(Uri::root(), 403);
		}

		list($invoicePath, $fileName) = PMFormHelper::generateInvoicePDF($row);

		$this->processDownloadFile($invoicePath, $fileName);
	}

	/**
	 * Validate form data
	 *
	 * @param   int    $formId
	 * @param   array  $data
	 *
	 * @return array
	 */
	protected function validateFormData($formId, $data)
	{
		$rowFields = PMFormHelper::getFormFields($formId);
		$form      = new \POSFForm($rowFields);
		$form->bind($data);
		$form->buildFieldsDependency(false);

		return $form->validate();
	}

	/**
	 * Method to validate username
	 *
	 * @param   string  $username
	 *
	 * @return array
	 */
	public function validateUsername($username)
	{
		$filterInput = \JFilterInput::getInstance();
		$db          = $this->container->db;
		$errors      = [];

		if (empty($username))
		{
			$errors[] = Text::sprintf('PF_FORM_FIELD_IS_REQURED', Text::_('PF_USERNAME'));
		}

		if ($filterInput->clean($username, 'TRIM') == '')
		{
			$errors[] = Text::_('JLIB_DATABASE_ERROR_PLEASE_ENTER_A_USER_NAME');
		}

		if (preg_match('#[<>"\'%;()&\\\\]|\\.\\./#', $username) || strlen(utf8_decode($username)) < 2
			|| $filterInput->clean($username, 'TRIM') !== $username
		)
		{
			$errors[] = Text::sprintf('JLIB_DATABASE_ERROR_VALID_AZ09', 2);
		}

		$query = $db->getQuery(true)
			->select('COUNT(*)')
			->from('#__users')
			->where('username = ' . $db->quote($username));
		$db->setQuery($query);
		$total = $db->loadResult();

		if ($total)
		{
			$errors[] = Text::_('PF_VALIDATION_INVALID_USERNAME');
		}

		return $errors;
	}

	/**
	 * Method to validate password
	 *
	 * @param   string  $password
	 *
	 * @return array
	 */
	public function validatePassword($password)
	{
		//Load language from user component
		$lang = Factory::getLanguage();
		$tag  = $lang->getTag();

		if (!$tag)
		{
			$tag = 'en-GB';
		}

		$lang->load('com_users', JPATH_ROOT, $tag);

		$errors = [];

		$params           = ComponentHelper::getParams('com_users');
		$minimumIntegers  = $params->get('minimum_integers');
		$minimumSymbols   = $params->get('minimum_symbols');
		$minimumUppercase = $params->get('minimum_uppercase');

		// We don't allow white space inside passwords
		$valueTrim   = trim($password);
		$valueLength = strlen($password);

		if (strlen($valueTrim) !== $valueLength)
		{
			$errors[] = Text::_('COM_USERS_MSG_SPACES_IN_PASSWORD');
		}

		if (!empty($minimumIntegers))
		{
			$nInts = preg_match_all('/[0-9]/', $password, $imatch);

			if ($nInts < $minimumIntegers)
			{
				$errors[] = Text::plural('COM_USERS_MSG_NOT_ENOUGH_INTEGERS_N', $minimumIntegers);
			}
		}

		if (!empty($minimumSymbols))
		{
			$nsymbols = preg_match_all('[\W]', $password, $smatch);

			if ($nsymbols < $minimumSymbols)
			{
				$errors[] = Text::plural('COM_USERS_MSG_NOT_ENOUGH_SYMBOLS_N', $minimumSymbols);
			}
		}

		if (!empty($minimumUppercase))
		{
			$nUppercase = preg_match_all("/[A-Z]/", $password, $umatch);

			if ($nUppercase < $minimumUppercase)
			{
				$errors[] = Text::plural('COM_USERS_MSG_NOT_ENOUGH_UPPERCASE_LETTERS_N', $minimumUppercase);
			}
		}

		return $errors;
	}

	/**
	 * Validate email for user account
	 *
	 * @param   string  $email
	 * @param   bool    $checkExists
	 * @param   int     $userId
	 *
	 * @return array
	 */
	protected function validateEmail($email)
	{
		$filterInput = \JFilterInput::getInstance();
		$errors      = [];

		// Validate email
		if (empty($email))
		{
			$errors[] = Text::sprintf('PF_FIELD_NAME_IS_REQUIRED', Text::_('Email'));
		}


		if (($filterInput->clean($email, 'TRIM') == "") || !MailHelper::isEmailAddress($email))
		{
			$errors[] = Text::_('JLIB_DATABASE_ERROR_VALID_MAIL');
		}

		$db    = $this->container->db;
		$query = $db->getQuery(true)
			->select('COUNT(*)')
			->from('#__users')
			->where('email = ' . $db->quote($email));

		$db->setQuery($query);

		$total = $db->loadResult();

		if ($total)
		{
			$errors[] = Text::_('PF_VALIDATION_INVALID_EMAIL');
		}

		return $errors;
	}
}