<?php
/**
 * @version        6.2.0
 * @package        Joomla
 * @subpackage     Payment Form
 * @author         Tuan Pham Ngoc
 * @copyright      Copyright (C) 2010 - 2021 Ossolution Team
 * @license        GNU/GPL, see LICENSE.php
 */

namespace OSSolution\PaymentForm\Site\Helper;

use JFormHelper;
use Joomla\CMS\Component\ComponentHelper;
use Joomla\CMS\Factory;
use Joomla\CMS\Filesystem\File;
use Joomla\CMS\Filesystem\Folder;
use Joomla\CMS\Form\Form;
use Joomla\CMS\HTML\HTMLHelper;
use Joomla\CMS\Http\HttpFactory;
use Joomla\CMS\Language\LanguageHelper;
use Joomla\CMS\Language\Multilanguage;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Mail\MailHelper;
use Joomla\CMS\Table\Table;
use Joomla\CMS\Uri\Uri;
use Joomla\Registry\Registry;
use Joomla\String\StringHelper;
use OSL\Container\Container;
use OSL\Utils\Database as DatabaseUtils;

defined('_JEXEC') or die;

class Helper
{
	/**
	 * Get current version number of Paymennt Form
	 *
	 * @return string
	 */
	public static function getInstalledVersion()
	{
		return '6.5.2';
	}

	/**
	 * Helper method to check if the extension is running on Joomla 4
	 *
	 * @return bool
	 */
	public static function isJoomla4()
	{
		return version_compare(JVERSION, '4.0.0-dev', 'ge');
	}

	/**
	 * Detect original filename of a uploaded file
	 *
	 * @param   string  $fileName
	 *
	 * @return string
	 */
	public static function getOriginalFilename($fileName)
	{
		$pos = strpos($fileName, '_');

		if ($pos !== false)
		{
			$fileName = substr($fileName, $pos + 1);
		}

		return $fileName;
	}

	/**
	 * Get configuration data and store in config object
	 *
	 * @return object
	 */
	public static function getConfig()
	{
		static $config;

		if (!$config)
		{
			$db     = Factory::getDbo();
			$query  = $db->getQuery(true);
			$config = new \stdClass();
			$query->select('*')
				->from('#__pf_configs');
			$db->setQuery($query);
			$rows = $db->loadObjectList();

			for ($i = 0, $n = count($rows); $i < $n; $i++)
			{
				$row          = $rows[$i];
				$key          = $row->config_key;
				$config->$key = stripslashes($row->config_value);
			}
		}

		return $config;
	}

	/**
	 * Get specify config value
	 *
	 * @param   string  $key
	 *
	 * @return string
	 */
	public static function getConfigValue($key)
	{
		$config = self::getConfig();

		return $config->{$key};
	}

	/**
	 * Prepare multilingual messages
	 *
	 * @param   \stdClass  $config
	 * @param   array      $keys
	 * @param   string     $prefix
	 *
	 * @return void
	 */
	public static function prepareMultilingualMessages($config, $keys, $prefix)
	{
		foreach ($keys as $key)
		{
			if (isset($config->{$key . $prefix}) && static::isValidMessage($config->{$key . $prefix}))
			{
				$config->{$key} = $config->{$key . $prefix};
			}
		}
	}

	/**
	 * Prepare multilingual messages
	 *
	 * @param   \stdClass  $config
	 * @param   array      $keys
	 * @param   string     $prefix
	 *
	 * @return void
	 */
	public static function prepareMultilingualStrings($config, $keys, $prefix)
	{
		foreach ($keys as $key)
		{
			if (isset($config->{$key . $prefix}) && strlen($config->{$key . $prefix}))
			{
				$config->{$key} = $config->{$key . $prefix};
			}
		}
	}

	/**
	 * Save config value into database
	 *
	 * @param   string  $key
	 * @param   string  $value
	 */
	public static function saveConfigValue($key, $value)
	{
		$db    = Factory::getDbo();
		$query = $db->getQuery(true);
		$query->update('#__pf_configs')
			->set('config_value = ' . $db->quote($value))
			->where('config_key = ' . $db->quote($key));
		$db->setQuery($query);
		$db->execute();
	}

	/**
	 * Method to call a static overridable helper method
	 *
	 * @param   string  $helper
	 * @param   string  $method
	 * @param   array   $methodArgs
	 *
	 * @return mixed
	 *
	 * @throws \Exception
	 */
	public static function callOverridableHelperMethod($helper, $method, $methodArgs = [])
	{
		$callableMethods   = [];
		$callableMethods[] = '\\OSSolution\\PaymentForm\\Site\\Override\\Helper\\' . ucfirst($helper) . '::' . $method;
		$callableMethods[] = '\\OSSolution\\PaymentForm\\Site\\Helper\\' . ucfirst($helper) . '::' . $method;


		foreach ($callableMethods as $callable)
		{
			if (is_callable($callable))
			{
				return call_user_func_array($callable, $methodArgs);
			}
		}

		throw new \Exception(sprintf('Method %s does not exist in the helper %s', $method, $helper));
	}

	/**
	 * Get Itemid of Payment Form comoonent
	 *
	 * @return int
	 */
	public static function getItemid()
	{
		$app       = Factory::getApplication();
		$menus     = $app->getMenu('site');
		$component = ComponentHelper::getComponent('com_pmform');
		$items     = $menus->getItems('component_id', $component->id);
		$views     = ['forms', 'form'];

		foreach ($views as $view)
		{
			$viewUrl = 'index.php?option=com_pmform&view=' . $view;

			foreach ($items as $item)
			{
				if (strpos($item->link, $viewUrl) !== false)
				{
					return $item->id;
				}
			}
		}

		return 0;
	}

	/**
	 * Get page params of the given view
	 *
	 * @param $active
	 * @param $views
	 *
	 * @return Registry
	 */
	public static function getViewParams($active, $views)
	{
		$views = (array) $views;

		if ($active && isset($active->query['view']) && in_array($active->query['view'], $views))
		{
			return $active->getParams();
		}

		return new Registry;
	}

	/**
	 * Get URL of the site, using for Ajax request
	 */
	public static function getSiteUrl()
	{
		return Uri::root();
	}

	/**
	 * Get BootstrapHelper class for admin UI
	 *
	 * @return \PMFormHelperBootstrap
	 */
	public static function getAdminBootstrapHelper()
	{
		if (static::isJoomla4())
		{
			return new \PMFormHelperBootstrap('5');
		}

		return new \PMFormHelperBootstrap('2');
	}

	/**
	 * Get list of fields used on the form
	 *
	 * @param   int     $formId
	 * @param   string  $fieldSuffix
	 *
	 * @return array
	 */
	public static function getFormFields($formId, $fieldSuffix = null)
	{
		$user  = Factory::getUser();
		$db    = Factory::getDbo();
		$query = $db->getQuery(true);
		$query->select('a.*')
			->from('#__pf_fields AS a')
			->where('a.published = 1')
			->where('(form_id = -1 OR id IN (SELECT field_id FROM #__pf_form_fields WHERE form_id=' . $formId . '))')
			->order('a.ordering');

		if (!$user->authorise('core.admin'))
		{
			$query->where('a.access IN (' . implode(',', $user->getAuthorisedViewLevels()) . ')');
		}

		if ($fieldSuffix)
		{
			DatabaseUtils::getMultilingualFields(
				$query,
				['title', 'description', 'options_text', 'place_holder', 'validation_error_message'],
				$fieldSuffix
			);
		}

		$db->setQuery($query);

		return $db->loadObjectList();
	}

	/**
	 * Get payment data
	 *
	 * @param   \OSSolution\PaymentForm\Admin\Table\Payment  $row
	 * @param   array                                        $rowFields
	 * @param   bool                                         $convertDate
	 * @param   bool                                         $output
	 *
	 * @return array
	 */
	public static function getPaymentData($row, $rowFields, $convertDate = true, $output = false)
	{
		$config      = self::getConfig();
		$db          = Factory::getDbo();
		$query       = $db->getQuery(true);
		$fieldSuffix = static::getFieldSuffix($row->language);

		$query->select('a.name, a.fieldtype, a.values, b.field_value')
			->select($db->quoteName('a.options_text' . $fieldSuffix, 'options_text'))
			->from('#__pf_fields AS a')
			->innerJoin('#__pf_field_value AS b ON a.id = b.field_id ')
			->where('payment_id=' . (int) $row->id);
		$db->setQuery($query);
		$fieldValues = $db->loadObjectList('name');

		$data = [];

		foreach ($rowFields as $rowField)
		{
			$fieldValue = null;

			if ($rowField->is_core)
			{
				$data[$rowField->name] = $row->{$rowField->name};
				$fieldValue            = $row->{$rowField->name};
			}
			else
			{
				if (isset($fieldValues[$rowField->name]))
				{
					$fieldValue            = $fieldValues[$rowField->name]->field_value;
					$data[$rowField->name] = $fieldValues[$rowField->name]->field_value;
				}
			}

			if ($output && $fieldValue && in_array($rowField->fieldtype, ['List', 'Checkboxes', 'Radio']))
			{
				$fieldValue = static::getTextValue($rowField, $fieldValue);
			}

			$data[$rowField->name] = $fieldValue;

			if ($convertDate && $rowField->fieldtype == 'Date' && !empty($data[$rowField->name]))
			{
				// Try to convert the format
				try
				{
					$date = new \DateTime($data[$rowField->name]);

					if ($date)
					{
						$data[$rowField->name] = $date->format($config->date_format);
					}
				}
				catch (\Exception $e)
				{

				}
			}
		}

		return $data;
	}

	/**
	 * Get text value for an option of the field
	 *
	 * @param   \stdClass  $rowField
	 * @param   string     $fieldValue
	 *
	 * @return string
	 */
	public static function getTextValue($rowField, $fieldValue)
	{
		$values      = explode("\r\n", $rowField->values);
		$optionsText = explode("\r\n", $rowField->options_text);
		$map         = [];

		for ($i = 0, $n = count($values); $i < $n; $i++)
		{
			if (isset($optionsText[$i]) && strlen($optionsText[$i]) > 0)
			{
				$map[$values[$i]] = $optionsText[$i];
			}
			else
			{
				$map[$values[$i]] = $values[$i];
			}
		}

		if (is_string($fieldValue) && is_array(json_decode($fieldValue)))
		{
			$fieldValues     = json_decode($fieldValue);
			$fieldValuesText = [];

			foreach ($fieldValues as $value)
			{
				if (isset($map[$value]))
				{
					$fieldValuesText[] = $map[$value];
				}
			}

			return json_encode($fieldValuesText);
		}
		elseif (isset($map[$fieldValue]))
		{
			return $map[$fieldValue];
		}
		else
		{
			return $fieldValue;
		}
	}

	/**
	 * Get all dependencies custom fields of a given field
	 *
	 * @param $id
	 *
	 * @return array
	 */
	public static function getAllDependencyFields($id)
	{
		$db    = Factory::getDbo();
		$query = $db->getQuery(true);

		$queue  = [$id];
		$fields = [$id];

		while (count($queue))
		{
			$masterFieldId = array_pop($queue);

			//Get list of dependency fields of this master field
			$query->clear()
				->select('id')
				->from('#__pf_fields')
				->where('depend_on_field_id = ' . $masterFieldId);
			$db->setQuery($query);
			$rows = $db->loadObjectList();

			if (count($rows))
			{
				foreach ($rows as $row)
				{
					$queue[]  = $row->id;
					$fields[] = $row->id;
				}
			}
		}

		return $fields;
	}

	/**
	 * Get Form data
	 *
	 * @param $rowFields
	 * @param $userId
	 * @param $config
	 *
	 * @return array
	 */
	public static function getFormData($rowFields, $userId, $config)
	{
		$data = [];

		if ($userId)
		{
			if ($config->cb_integration == 1)
			{
				$synchronizer = new \POSFSynchronizerCommunitybuilder();
				$mappings     = [];

				foreach ($rowFields as $rowField)
				{
					if ($rowField->field_mapping)
					{
						$mappings[$rowField->name] = $rowField->field_mapping;
					}
				}

				$data = $synchronizer->getData($userId, $mappings);
			}
			elseif ($config->cb_integration == 2)
			{
				$synchronizer = new \POSFSynchronizerJomsocial();
				$mappings     = [];

				foreach ($rowFields as $rowField)
				{
					if ($rowField->field_mapping)
					{
						$mappings[$rowField->name] = $rowField->field_mapping;
					}
				}

				$data = $synchronizer->getData($userId, $mappings);
			}
			elseif ($config->cb_integration == 3)
			{
				$synchronizer = new \POSFSynchronizerJoomla();
				$mappings     = [];

				foreach ($rowFields as $rowField)
				{
					if ($rowField->field_mapping)
					{
						$mappings[$rowField->name] = $rowField->field_mapping;
					}
				}

				$data = $synchronizer->getData($userId, $mappings);
			}
			else
			{
				$db    = Factory::getDbo();
				$query = $db->getQuery(true);
				$query->select('*')
					->from('#__pf_payments')
					->where('user_id=' . (int) $userId)
					->order('id DESC');
				$db->setQuery($query, 0, 1);
				$row = $db->loadObject();

				if ($row)
				{
					$data = self::getPaymentData($row, $rowFields);
				}
			}
		}

		return $data;
	}


	/**
	 * Store custom fields data
	 *
	 * @param $row
	 * @param $data
	 */
	public static function storeFormData($row, $data)
	{
		$container = Container::getInstance('com_pmform');

		$db    = $container->db;
		$query = $db->getQuery(true);

		$rowFieldValue = $container->factory->createTable('Fieldvalue');
		$fieldSuffix   = static::getFieldSuffix($row->language);

		$query->delete('#__pf_field_value')->where('payment_id=' . (int) $row->id);
		$db->setQuery($query);
		$db->execute();

		$config       = self::getConfig();
		$uploadFolder = $config->path_upload;

		if (!$uploadFolder)
		{
			$uploadFolder = 'images/com_pmform';
		}

		$pathUpload = JPATH_ROOT . '/' . $uploadFolder;

		if (!Folder::exists($pathUpload))
		{
			Folder::create($pathUpload);
		}

		$allowedExtensions = $config->allowed_extensions;

		if (!$allowedExtensions)
		{
			$allowedExtensions = 'doc, docx, ppt, pptx, pdf, zip, rar, jpg, jepg, png, zipx';
		}

		$allowedExtensions = explode(',', $allowedExtensions);
		$allowedExtensions = array_map('trim', $allowedExtensions);

		// Date format
		$dateFormat = $config->date_field_format ? $config->date_field_format : '%Y-%m-%d';
		$dateFormat = str_replace('%', '', $dateFormat);

		$rowFields = self::getFormFields($row->form_id, $fieldSuffix);

		foreach ($rowFields as $rowField)
		{
			if ($rowField->is_core)
			{
				continue;
			}

			if ($rowField->fieldtype == 'File')
			{
				$name = $rowField->name;

				// If there are field, we need to upload the file to server and save it !
				if (isset($_FILES[$name]))
				{
					if ($_FILES[$name]['name'] != '')
					{
						$fileName = $_FILES[$name]['name'];
						$fileExt  = File::getExt($fileName);

						if (in_array(strtolower($fileExt), $allowedExtensions))
						{
							$fileName = File::makeSafe($fileName);
							if (File::exists($pathUpload . '/' . $fileName))
							{
								$targetFileName = time() . '_' . $fileName;
							}
							else
							{
								$targetFileName = $fileName;
							}

							File::upload($_FILES[$name]['tmp_name'], $pathUpload . '/' . $targetFileName);

							$data[$name] = $targetFileName;
						}
					}
				}
			}

			if ($rowField->fieldtype == 'Date')
			{
				$fieldValue = $data[$rowField->name];

				if ($fieldValue)
				{
					// Try to convert the format
					try
					{
						$date = \DateTime::createFromFormat($dateFormat, $fieldValue);

						if ($date)
						{
							$fieldValue = $date->format('Y-m-d');
						}
						else
						{
							$fieldValue = '';
						}
					}
					catch (\Exception $e)
					{
						$fieldValue = '';
					}

					$data[$rowField->name] = $fieldValue;
				}
			}

			$fieldValue                = isset($data[$rowField->name]) ? $data[$rowField->name] : null;
			$rowFieldValue->id         = 0;
			$rowFieldValue->field_id   = $rowField->id;
			$rowFieldValue->payment_id = $row->id;

			if (is_array($fieldValue))
			{
				$fieldValue = json_encode($fieldValue);
			}

			$rowFieldValue->field_value = $fieldValue;

			$rowFieldValue->field_value_text = static::getTextValue($rowField, $fieldValue);

			$rowFieldValue->store();
		}
	}

	/**
	 * Check if the payment method is enabled or not
	 *
	 * @param $paymentMethod string Name of payment method
	 *
	 * @return bool True if the payment method is enabled, otherwise false
	 */
	public static function isPaymentMethodEnabled($paymentMethod)
	{
		$db    = Factory::getDbo();
		$query = $db->getQuery(true);
		$query->select('COUNT(*)')
			->from('#__pf_payment_plugins')
			->where('published = 1')
			->where('name=' . $db->quote($paymentMethod));
		$db->setQuery($query);
		$total = (int) $db->loadResult();

		if ($total)
		{
			if ($paymentMethod == 'os_ideal')
			{
				require_once JPATH_ROOT . '/components/com_pmform/payments/ideal/ideal.class.php';
			}

			return true;
		}

		return false;
	}

	/**
	 * Display list of attachment
	 *
	 * @param   string  $attachment
	 * @param   object  $config
	 *
	 * @return string
	 */
	public static function attachmentList($attachment, $config)
	{
		$path      = JPATH_ROOT . '/media/com_pmform';
		$files     = Folder::files($path, strlen(trim($config->allowed_extensions)) ? str_replace(',', '|', $config->allowed_extensions) : 'bmp|gif|jpg|png|swf|zip|doc|pdf|xls');
		$options   = [];
		$options[] = HTMLHelper::_('select.option', '', Text::_('Select Attachment'));

		for ($i = 0, $n = count($files); $i < $n; $i++)
		{
			$file      = $files[$i];
			$options[] = HTMLHelper::_('select.option', $file, $file);
		}

		return HTMLHelper::_('select.genericlist', $options, 'attachment', 'class="form-select"', 'value', 'text', $attachment);
	}

	/**
	 * Load language from main component
	 *
	 */
	public static function loadLanguage()
	{
		static $loaded;

		if (!$loaded)
		{
			$lang = Factory::getLanguage();
			$tag  = $lang->getTag();

			if (!$tag)
			{
				$tag = 'en-GB';
			}

			$lang->load('com_pmform', JPATH_ROOT, $tag);
			$loaded = true;
		}
	}

	/**
	 * Get lang to append to an URL
	 *
	 * @return string
	 */
	public static function getLangLink()
	{
		$langLink = '';

		if (Multilanguage::isEnabled())
		{
			$languages = LanguageHelper::getLanguages('lang_code');
			$langLink  = '&lang=' . $languages[Factory::getLanguage()->getTag()]->sef;
		}

		return $langLink;
	}

	/**
	 * Create an user account based on the entered data
	 *
	 * @param   array  $data
	 *
	 * @return int
	 */

	public static function saveRegistration($data)
	{
		//Need to load com_users language file
		$lang = Factory::getLanguage();
		$tag  = $lang->getTag();

		if (!$tag)
		{
			$tag = 'en-GB';
		}

		$lang->load('com_users', JPATH_ROOT, $tag);
		$data['name']     = $data['first_name'] . ' ' . $data['last_name'];
		$data['password'] = $data['password2'] = $data['password1'];
		$data['email1']   = $data['email2'] = $data['email'];

		if (self::isJoomla4())
		{
			Form::addFormPath(JPATH_ROOT . '/components/com_users/forms');

			/* @var \Joomla\Component\Users\Site\Model\RegistrationModel $model */
			$model = Factory::getApplication()->bootComponent('com_users')
				->getMVCFactory()->createModel('Registration', 'Site', ['ignore_request' => true]);
		}
		else
		{
			require_once JPATH_ROOT . '/components/com_users/models/registration.php';

			if (Multilanguage::isEnabled())
			{
				Form::addFormPath(JPATH_ROOT . '/components/com_users/models/forms');
				Form::addFieldPath(JPATH_ROOT . '/components/com_users/models/fields');
			}

			$model = new \UsersModelRegistration;
		}

		$model->register($data);

		$db    = Factory::getDbo();
		$query = $db->getQuery(true);
		$query->select('id')
			->from('#__users')
			->where('username=' . $db->quote($data['username']));
		$db->setQuery($query);

		return (int) $db->loadResult();
	}

	/**
	 * Calculate the fees on the form
	 *
	 * @param   \stdClass  $rowForm
	 * @param   \POSFForm  $form
	 * @param              $input
	 *
	 * @return array
	 */
	public static function calculateFormFee($rowForm, $form, $input, $paymentMethod = null)
	{
		$db                = Factory::getDbo();
		$query             = $db->getQuery(true);
		$fees              = [];
		$fees['coupon_id'] = 0;
		$form->setAmount($rowForm->amount);
		$extraFee    = $form->calculateFee();
		$totalAmount = $rowForm->amount + $extraFee;

		//Calculate late fee
		if ($rowForm->number_days > 0 && ($rowForm->late_payment_date != $db->getNullDate()))
		{
			if ($rowForm->late_amount_type == 0)
			{
				$lateFee = round($totalAmount * $rowForm->late_payment_amount / 100, 2);
			}
			else
			{
				$lateFee = $rowForm->late_payment_amount;
			}

			$totalAmount += $lateFee;
		}

		$discountAmount = 0;
		$formId         = $input->getInt('form_id', 0);
		$couponCode     = $input->post->getString('coupon_code', '');

		if ($couponCode)
		{
			$query->clear()
				->select('*')
				->from('#__pf_coupons')
				->where('published=1')
				->where('code=' . $db->quote($couponCode) . '')
				->where('(valid_from="0000-00-00" OR valid_from <= NOW())')
				->where('(valid_to="0000-00-00" OR valid_to >= NOW())')
				->where('(times = 0 OR times > used)')
				->where('(form_id=0 OR form_id=' . $formId . ')');
			$db->setQuery($query);
			$coupon = $db->loadObject();

			if ($coupon)
			{
				if ($coupon->coupon_type == 0)
				{
					$discountAmount = $discountAmount + $totalAmount * $coupon->discount / 100;
				}
				else
				{
					$discountAmount = $discountAmount + $coupon->discount;
				}

				$couponValid       = 1;
				$fees['coupon_id'] = (int) $coupon->id;
			}
			else
			{
				$couponValid = 0;
			}
		}
		else
		{
			$couponValid = 1;
		}

		// Calculate tax rate
		if ($rowForm->tax_rate > 0 && ($totalAmount - $discountAmount) > 0)
		{

			$taxAmount = round(($totalAmount - $discountAmount) * $rowForm->tax_rate / 100, 2);
		}
		else
		{
			$taxAmount = 0;
		}

		$grossAmount = $totalAmount - $discountAmount + $taxAmount;

		// Calculate payment fee
		$paymentFeeAmount  = 0;
		$paymentFeePercent = 0;

		if ($paymentMethod)
		{
			$method            = PaymentMethods::loadPaymentMethod($paymentMethod);
			$params            = new Registry($method->params);
			$paymentFeeAmount  = (float) $params->get('payment_fee_amount');
			$paymentFeePercent = (float) $params->get('payment_fee_percent');
		}

		if (($paymentFeeAmount > 0 || $paymentFeePercent > 0) && $grossAmount > 0)
		{
			$fees['payment_processing_fee'] = round($paymentFeeAmount + $grossAmount * $paymentFeePercent / 100, 2);
			$grossAmount                    += $fees['payment_processing_fee'];
		}
		else
		{
			$fees['payment_processing_fee'] = 0;
		}

		$fees['total_amount']    = $totalAmount;
		$fees['discount_amount'] = $discountAmount;
		$fees['tax_amount']      = $taxAmount;
		$fees['gross_amount']    = $grossAmount;
		$fees['coupon_valid']    = $couponValid;

		return $fees;
	}

	/**
	 * Format the amount based on config settings
	 *
	 * @param $amount
	 * @param $config
	 *
	 * @return string
	 */
	public static function formatAmount($amount, $config)
	{
		return number_format($amount, 2);
	}

	/**
	 * Display amount with the currency symbol
	 *
	 * @param $amount
	 * @param $config
	 * @param $currencySymbol
	 *
	 * @return string
	 */
	public static function formatCurrency($amount, $config, $currencySymbol = null)
	{
		$currencySymbol = $currencySymbol ?: $config->currency_symbol;

		return $currencySymbol . number_format($amount, 2);
	}

	/**
	 * Get country code
	 *
	 * @param   string  $countryName
	 *
	 * @return string
	 */
	public static function getCountryCode($countryName)
	{
		$db    = Factory::getDbo();
		$query = $db->getQuery(true);

		if (!$countryName)
		{
			$countryName = self::getConfigValue('default_country');
		}

		$query->select('country_2_code')
			->from('#__pf_countries')
			->where('LOWER(name) = ' . $db->quote(StringHelper::strtolower($countryName)));
		$db->setQuery($query);
		$countryCode = $db->loadResult();

		if (!$countryCode)
		{
			$countryCode = 'US';
		}

		return $countryCode;
	}

	/**
	 * This function is used to check to see whether we need to update the database to support multilingual or not
	 *
	 * @return boolean
	 */
	public static function isSynchronized()
	{
		$db             = Factory::getDbo();
		$fields         = array_keys($db->getTableColumns('#__pf_forms'));
		$extraLanguages = self::getLanguages();

		if (count($extraLanguages))
		{
			foreach ($extraLanguages as $extraLanguage)
			{
				$prefix = $extraLanguage->sef;

				if (!in_array('title_' . $prefix, $fields))
				{
					return false;
				}
			}
		}

		return true;
	}

	/**
	 * Synchronize Payment Form database to support multilingual
	 */
	public static function setupMultilingual()
	{
		$db        = Factory::getDbo();
		$languages = self::getLanguages();

		if (!count($languages))
		{
			return;
		}

		$formTableFields  = array_keys($db->getTableColumns('#__pf_forms'));
		$fieldTableFields = array_keys($db->getTableColumns('#__pf_fields'));

		foreach ($languages as $language)
		{
			#Process for #__pf_forms table
			$prefix = $language->sef;

			$varcharFields = [
				'title',
				'notification_email_subject',
				'confirmation_email_subject',
			];

			foreach ($varcharFields as $varcharField)
			{
				$fieldName = $varcharField . '_' . $prefix;

				if (!in_array($fieldName, $formTableFields))
				{
					$sql = "ALTER TABLE  `#__pf_forms` ADD  `$fieldName` VARCHAR( 255 );";
					$db->setQuery($sql);
					$db->execute();
				}
			}

			$textFields = [
				'form_msg',
				'confirmation_msg',
				'thanks_message',
				'cancel_message',
				'notification_email_body',
				'confirmation_email_body',
				'user_email_body_offline',
				'thanks_message_offline',
			];

			foreach ($textFields as $textField)
			{
				$fieldName = $textField . '_' . $prefix;

				if (!in_array($fieldName, $formTableFields))
				{
					$sql = "ALTER TABLE  `#__pf_forms` ADD  `$fieldName` TEXT NULL;";
					$db->setQuery($sql);
					$db->execute();
				}
			}

			// Process for #__pf_fields table

			$varcharFields = [
				'title',
				'place_holder',
				'validation_error_message',
			];

			foreach ($varcharFields as $varcharField)
			{
				$fieldName = $varcharField . '_' . $prefix;

				if (!in_array($fieldName, $fieldTableFields))
				{
					$sql = "ALTER TABLE  `#__pf_fields` ADD  `$fieldName` VARCHAR( 255 );";
					$db->setQuery($sql);
					$db->execute();
				}
			}

			$textFields = [
				'description',
				'options_text',
			];

			foreach ($textFields as $textField)
			{
				$fieldName = $textField . '_' . $prefix;

				if (!in_array($fieldName, $fieldTableFields))
				{
					$sql = "ALTER TABLE  `#__pf_fields` ADD  `$fieldName` TEXT NULL;";
					$db->setQuery($sql);
					$db->execute();
				}
			}
		}
	}

	/**
	 *
	 * Function to get all available languages except the default language
	 * @return array object list
	 */
	public static function getLanguages()
	{
		$languages = LanguageHelper::getLanguages('lang_code');
		unset($languages[self::getDefaultLanguage()]);

		return array_values($languages);
	}

	/**
	 * Get front-end default language
	 * @return string
	 */
	public static function getDefaultLanguage()
	{
		$params = ComponentHelper::getParams('com_languages');

		return $params->get('site', 'en-GB');
	}

	/**
	 * Get field suffix used in sql query
	 *
	 * @param   string  $activeLanguage
	 *
	 * @return string
	 */
	public static function getFieldSuffix($activeLanguage = null)
	{
		$prefix = '';

		if ($activeLanguage !== '*' && Multilanguage::isEnabled())
		{
			if (!$activeLanguage)
			{
				$activeLanguage = Factory::getLanguage()->getTag();
			}

			if ($activeLanguage != self::getDefaultLanguage())
			{
				$languages = LanguageHelper::getLanguages('lang_code');

				if (isset($languages[$activeLanguage]))
				{
					$prefix = '_' . $languages[$activeLanguage]->sef;
				}
			}
		}

		return $prefix;
	}

	/**
	 * Display copy right information from back-end of the extension
	 *
	 */
	public static function displayCopyRight()
	{
		echo '<div class="copyright"><a href="https://www.joomdonation.com/joomla-extensions/joomla-payment-form.html" target="_blank"><strong>Payment Form</strong></a> version ' . self::getInstalledVersion() . ', Copyright (C) 2010 - ' . date('Y') . ' <a href="http://joomdonation.com" target="_blank"><strong>Ossolution Team</strong></a></div>';
	}

	/**
	 * Build replacement tags use for emails & messages
	 *
	 * @param $row
	 * @param $config
	 *
	 * @return array
	 */
	public static function buildReplaceTags($row, $config, $loadCss = false)
	{
		$db          = Factory::getDbo();
		$query       = $db->getQuery(true);
		$fieldSuffix = static::getFieldSuffix($row->language);
		$query->select('*')
			->from('#__pf_forms')
			->where('id = ' . (int) $row->form_id);

		if ($fieldSuffix)
		{
			DatabaseUtils::getMultilingualFields($query, ['title'], $fieldSuffix);
		}

		$db->setQuery($query);
		$form = $db->loadObject();

		// Override currency symbol from form
		if ($form->currency_symbol)
		{
			$config->currency_symbol = $form->currency_symbol;
		}

		$rowFields                          = self::getFormFields($row->form_id, $fieldSuffix);
		$replaces                           = [];
		$replaces['form_title']             = $form->title;
		$replaces['form_id']                = $form->id;
		$replaces['id']                     = $row->id;
		$replaces['amount']                 = $replaces['total_amount'] = self::formatAmount($row->total_amount, $config);
		$replaces['discount_amount']        = self::formatAmount($row->discount_amount, $config);
		$replaces['tax_amount']             = self::formatAmount($row->tax_amount, $config);
		$replaces['gross_amount']           = self::formatAmount($row->gross_amount, $config);
		$replaces['payment_processing_fee'] = self::formatAmount($row->payment_processing_fee, $config);
		$replaces['transaction_id']         = $row->transaction_id;
		$replaces['date']                   = date($config->date_format);
		$replaces['payment_date']           = HTMLHelper::_('date', $row->created_date, $config->date_format);
		$replaces['payment_detail']         = self::getEmailContent($config, $row, $loadCss);
		$replaces['invoice_number']         = self::formatInvoiceNumber($row->invoice_number, $config);
		$method                             = PaymentMethods::getPaymentMethod($row->payment_method);

		if ($method)
		{
			$replaces['payment_method'] = Text::_($method->getTitle());
		}
		else
		{
			$replaces['payment_method'] = '';
		}

		$replaces = array_merge($replaces, self::getPaymentData($row, $rowFields, true, true));

		foreach ($rowFields as $rowField)
		{
			if (isset($replaces[$rowField->name]))
			{
				continue;
			}

			if ($rowField->fieldtype == 'Heading')
			{
				$replaces[$rowField->name] = $rowField->title;
			}

			if ($rowField->fieldtype == 'Message')
			{
				$replaces[$rowField->name] = $rowField->description;
			}
		}

		if ($row->coupon_id > 0)
		{
			$query->clear()
				->select('code')
				->from('#__pf_coupons')
				->where('id = ' . $row->coupon_id);
			$db->setQuery($query);
			$replaces['coupon_code'] = $db->loadResult();
		}
		else
		{
			$replaces['coupon_code'] = '';
		}

		// Make sure if a custom field is not available, the used tag would be empty
		$query->clear()
			->select('name')
			->from('#__pf_fields')
			->where('published = 1');
		$db->setQuery($query);
		$allFields = $db->loadColumn();

		foreach ($allFields as $field)
		{
			if (!isset($replaces[$field]))
			{
				$replaces[$field] = '';
			}
		}

		$replaces['voucher_code'] = '';
		$replaces['ip_address']   = $row->ip_address;

		if ($row->eventsbooking_coupon_id > 0)
		{
			$query->clear()
				->select($db->quoteName('code'))
				->from('#__eb_coupons')
				->where('id = ' . $row->eventsbooking_coupon_id);
			$db->setQuery($query);
			$replaces['voucher_code'] = $db->loadResult();
		}

		// Support getting form titles from none default languages
		$languages = self::getLanguages();

		foreach ($languages as $language)
		{
			$sef                            = $language->sef;
			$replaces['form_title_' . $sef] = $form->{'title_' . $sef};
		}

		return $replaces;
	}

	/**
	 * Send email to super administrator and user
	 *
	 * @param   \OSSolution\PaymentForm\Admin\Table\Payment  $row
	 * @param   \stdClass                                    $config
	 */
	public static function sendEmails($row, $config)
	{
		$db     = Factory::getDbo();
		$mailer = static::getMailer();

		// Load language if needed
		if (Factory::getApplication()->isClient('administrator') && $row->language && $row->language != '*')
		{
			Factory::getLanguage()->load('com_pmform', JPATH_ROOT, $row->language);
		}

		$fieldSuffix = static::getFieldSuffix($row->language);
		$rowFields   = self::getFormFields((int) $row->form_id, $fieldSuffix);

		$query = $db->getQuery(true);
		$query->select('*')
			->from('#__pf_forms')
			->where('id = ' . (int) $row->form_id);
		$db->setQuery($query);
		$form = $db->loadObject();

		if ($fieldSuffix)
		{
			$formKeys = [
				'user_email_body_offline',
				'confirmation_email_body',
				'notification_email_body',
			];

			static::prepareMultilingualMessages($form, $formKeys, $fieldSuffix);

			$formKeys = [
				'confirmation_email_subject',
				'notification_email_subject',
			];

			static::prepareMultilingualStrings($form, $formKeys, $fieldSuffix);

			$configKeys = [
				'user_email_body_offline',
				'user_email_body',
				'admin_email_body',
			];

			static::prepareMultilingualMessages($config, $configKeys, $fieldSuffix);

			$configKeys = [
				'user_email_subject',
				'admin_email_subject',
			];

			static::prepareMultilingualStrings($config, $configKeys, $fieldSuffix);
		}

		$replaces = self::callOverridableHelperMethod('Helper', 'buildReplaceTags', [$row, $config, true]);

		// Prepare messages in correct languages

		if ($form->confirmation_email_subject)
		{
			$config->user_email_subject = $form->confirmation_email_subject;
		}

		if ($form->notification_emails)
		{
			$config->notification_emails = $form->notification_emails;
		}

		//Notification email send to user
		$subject = $config->user_email_subject;

		if (strpos($row->payment_method, 'os_offline') !== false)
		{
			$prefix = str_replace('os_offline', '', $row->payment_method);

			if (self::isValidMessage($form->user_email_body_offline))
			{
				$body = $form->user_email_body_offline;
			}
			elseif ($prefix && self::isValidMessage($config->{'user_email_body_offline' . $prefix}))
			{
				$body = $config->{'user_email_body_offline' . $prefix};
			}
			else
			{
				$body = $config->user_email_body_offline;
			}
		}
		else
		{
			if (self::isValidMessage($form->confirmation_email_body))
			{
				$body = $form->confirmation_email_body;
			}
			else
			{
				$body = $config->user_email_body;
			}
		}

		foreach ($replaces as $key => $value)
		{
			$key     = strtoupper($key);
			$subject = str_replace("[$key]", $value, $subject);
			$body    = str_replace("[$key]", $value, $body);
		}

		$body = self::convertImgTags($body);

		$attachments = [];
		$invoicePath = '';

		if ($config->activate_invoice_feature && self::needInvoice($row))
		{
			if (!$row->invoice_number)
			{
				$row->invoice_number = self::getInvoiceNumber();
				$row->store();
			}

			self::generateInvoicePDF($row);

			$invoicePath = JPATH_ROOT . '/media/com_pmform/invoices/' . self::formatInvoiceNumber($row->invoice_number, $config) . '.pdf';
		}

		if ($config->send_invoice_to_customer && $invoicePath)
		{
			$attachments[] = $invoicePath;
		}

		if ($form->attachment)
		{
			$attachments[] = JPATH_ROOT . '/media/com_pmform/' . $form->attachment;
		}

		if (count($attachments))
		{
			$mailer->addAttachment($attachments);
		}

		if (MailHelper::isEmailAddress($row->email))
		{
			foreach ($rowFields as $rowField)
			{
				if ($rowField->receive_confirmation_email && !empty($replaces[$rowField->name]) && MailHelper::isEmailAddress($replaces[$rowField->name]))
				{
					$mailer->addBcc($replaces[$rowField->name]);
				}
			}

			reset($rowFields);

			static::send($mailer, [$row->email], $subject, $body);
		}

		$mailer->clearAllRecipients();

		if (count($attachments))
		{
			$mailer->clearAttachments();
		}

		//Send emails to notification emails
		$replaces['payment_detail'] = self::getEmailContent($config, $row, true, true);

		$emails = explode(',', $config->notification_emails);
		$emails = array_map('trim', $emails);

		if ($form->notification_email_subject)
		{
			$subject = $form->notification_email_subject;
		}
		else
		{
			$subject = $config->admin_email_subject;
		}

		if (static::isValidMessage($form->notification_email_body))
		{
			$body = $form->notification_email_body;
		}
		else
		{
			$body = $config->admin_email_body;
		}

		foreach ($replaces as $key => $value)
		{
			$key     = strtoupper($key);
			$subject = str_replace("[$key]", $value, $subject);
			$body    = str_replace("[$key]", $value, $body);
		}

		$body = self::convertImgTags($body);

		//Send attachments to admin emails
		if ($config->send_attachment_to_admin_email)
		{
			$pathUpload      = self::getConfigValue('path_upload');
			$attachmentsPath = JPATH_ROOT . '/' . $pathUpload;

			for ($i = 0, $n = count($rowFields); $i < $n; $i++)
			{
				$rowField = $rowFields[$i];

				if ($rowField->fieldtype == 'File')
				{
					if (isset($replaces[$rowField->name]))
					{
						$fileName = $replaces[$rowField->name];

						if ($fileName && file_exists($attachmentsPath . '/' . $fileName))
						{
							$pos = strpos($fileName, '_');

							if ($pos !== false)
							{
								$originalFilename = substr($fileName, $pos + 1);
							}
							else
							{
								$originalFilename = $fileName;
							}

							$mailer->addAttachment($attachmentsPath . '/' . $fileName, $originalFilename);
						}
					}
				}
			}
		}

		if ($config->send_invoice_to_admin && $invoicePath)
		{
			$mailer->addAttachment($invoicePath);
		}

		$mailer->addReplyTo($row->email);

		static::send($mailer, $emails, $subject, $body);

		//After sending email, we can empty the user_password of subscription was activated
		if ($row->published == 1 && $row->user_password)
		{
			$query->clear()
				->update('#__pf_payments')
				->set('user_password = ""')
				->where('id = ' . $row->id);
			$db->setQuery($query);
			$db->execute();
		}
	}

	/**
	 * Check if the given message entered via HTML editor has actual data
	 *
	 * @param $string
	 *
	 * @return bool
	 */
	public static function isValidMessage($string)
	{
		$string = strip_tags($string, '<img>');

		// Remove all special characters
		$string = str_replace(['.', ' ', "\n", "\t", "\r"], '', $string);

		$string = trim($string);

		if (strlen($string) > 10)
		{
			return true;
		}

		return false;
	}

	public static function sendPaymentApprovedEmails($row, $config)
	{
		if (!MailHelper::isEmailAddress($row->email))
		{
			return;
		}

		// Load language if needed
		if (Factory::getApplication()->isClient('administrator') && $row->language && $row->language != '*')
		{
			Factory::getLanguage()->load('com_pmform', JPATH_ROOT, $row->language);
		}

		$mailer = static::getMailer();

		$replaces = self::callOverridableHelperMethod('Helper', 'buildReplaceTags', [$row, $config, true]);

		$fieldSuffix = static::getFieldSuffix($row->language);

		if ($fieldSuffix)
		{
			static::prepareMultilingualMessages($config, ['payment_approved_email_body'], $fieldSuffix);
			static::prepareMultilingualStrings($config, ['payment_approved_email_subject'], $fieldSuffix);
		}

		//Notification email send to user
		$subject = $config->payment_approved_email_subject;
		$body    = $config->payment_approved_email_body;

		foreach ($replaces as $key => $value)
		{
			$key     = strtoupper($key);
			$subject = str_replace("[$key]", $value, $subject);
			$body    = str_replace("[$key]", $value, $body);
		}

		$body = self::convertImgTags($body);

		// Generate and
		if ($config->activate_invoice_feature && $config->send_invoice_to_customer && self::needInvoice($row))
		{
			if (!$row->invoice_number)
			{
				$row->invoice_number = self::getInvoiceNumber();
				$row->store();
			}

			self::generateInvoicePDF($row);

			$mailer->addAttachment(JPATH_ROOT . '/media/com_pmform/invoices/' . self::formatInvoiceNumber($row->invoice_number, $config) . '.pdf');
		}

		static::send($mailer, [$row->email], $subject, $body);

		//After sending email, we can empty the user_password of subscription was activated
		if ($row->published == 1 && $row->user_password)
		{
			$db    = Factory::getDbo();
			$query = $db->getQuery(true);
			$query->update('#__pf_payments')
				->set('user_password = ""')
				->where('id = ' . $row->id);
			$db->setQuery($query);
			$db->execute();
		}
	}

	/**
	 * Get email content. For [PAYMENT_DETAIL] tag
	 *
	 * @param   object  $config
	 * @param   object  $row
	 * @param   bool    $loadCss
	 * * @param   bool  $toAdmin
	 *
	 * @return string
	 */
	public static function getEmailContent($config, $row, $loadCss = false, $toAdmin = false)
	{
		$db          = Factory::getDbo();
		$query       = $db->getQuery(true);
		$fieldSuffix = static::getFieldSuffix($row->language);
		$query->select($db->quoteName('title' . $fieldSuffix))
			->from('#__pf_forms')
			->where('id = ' . (int) $row->form_id);
		$db->setQuery($query);
		$data              = [];
		$data['config']    = $config;
		$data['row']       = $row;
		$data['formTitle'] = $db->loadResult();

		//Get coupon
		if ($row->coupon_id)
		{
			$query->clear();
			$query->select('`code`')
				->from('#__pf_coupons')
				->where('id = ' . (int) $row->coupon_id);
			$db->setQuery($query);
			$data['couponCode'] = $db->loadResult();
		}

		//Username and password of the account
		if ($row->user_id)
		{
			$query->clear();
			$query->select('username')
				->from('#__users')
				->where('id = ' . $row->user_id);
			$db->setQuery($query);
			$username         = $db->loadResult();
			$data['username'] = $username;
		}

		if ($row->username && $row->user_password)
		{
			$data['username'] = $row->username;

			//Password
			$privateKey       = md5(Factory::getApplication()->get('secret'));
			$key              = new \JCryptKey('simple', $privateKey, $privateKey);
			$crypt            = new \JCrypt(new \JCryptCipherSimple, $key);
			$data['password'] = $crypt->decrypt($row->user_password);
		}

		$params = ComponentHelper::getParams('com_users');

		if (!$params->get('sendpassword', 1) && isset($data['password']))
		{
			unset($data['password']);
		}

		if ($loadCss)
		{
			$layout      = 'emailtemplates/tmpl/email.php';
			$data['css'] = file_get_contents(JPATH_ROOT . '/media/com_pmform/assets/css/style.min.css');
		}
		else
		{
			$layout = 'emailtemplates/tmpl/payment_detail.php';
		}

		$rowFields = self::getFormFields($row->form_id, $fieldSuffix);
		$formData  = self::getPaymentData($row, $rowFields, true, false);
		$form      = new \POSFForm($rowFields);
		$form->bind($formData)->prepareFormField();
		$form->buildFieldsDependency();
		$fields = $form->getFields();

		foreach ($fields as $field)
		{
			if (!$field->visible)
			{
				unset($fields[$field->name]);
			}
		}

		$form->setFields($fields);

		$data['form']        = $form;
		$data['enterAmount'] = self::isEnterAmount($rowFields);
		$data['toAdmin']     = $toAdmin;

		if ($row->eventsbooking_coupon_id > 0)
		{
			$query->clear()
				->select($db->quoteName('code'))
				->from('#__eb_coupons')
				->where('id = ' . $row->eventsbooking_coupon_id);
			$db->setQuery($query);
			$data['voucherCode'] = $db->loadResult();
		}

		return self::loadCommonLayout($layout, $data);
	}

	/**
	 * Convert payment amount to USD currency in case the currency is not supported by the payment gateway
	 *
	 * @param $amount
	 * @param $currency
	 *
	 * @return float
	 */
	public static function convertAmountToUSD($amount, $currency)
	{
		static $rate = null;

		if ($rate === null)
		{
			$http     = HttpFactory::getHttp();
			$url      = 'http://download.finance.yahoo.com/d/quotes.csv?e=.csv&f=sl1d1t1&s=USD' . $currency . '=X';
			$response = $http->get($url);

			if ($response->code == 200)
			{
				$currencyData = explode(',', $response->body);
				$rate         = floatval($currencyData[1]);
			}
		}

		if ($rate > 0)
		{
			$amount = $amount / $rate;
		}

		return round($amount, 2);
	}

	/**
	 * Export payment records into CSV file
	 *
	 * @param $rows
	 * @param $config
	 * @param $rowFields
	 * @param $fieldValues
	 *
	 * @throws \Exception
	 */
	public static function csvExport($rows, $config, $rowFields, $fieldValues)
	{
		if (count($rows))
		{
			if (preg_match('#Opera(/| )([0-9].[0-9]{1,2})#i', $_SERVER['HTTP_USER_AGENT']))
			{
				$UserBrowser = "Opera";
			}
			elseif (preg_match('#MSIE ([0-9].[0-9]{1,2})#i', $_SERVER['HTTP_USER_AGENT']))
			{
				$UserBrowser = "IE";
			}
			else
			{
				$UserBrowser = '';
			}

			$mime_type = ($UserBrowser == 'IE' || $UserBrowser == 'Opera') ? 'application/octetstream' : 'application/octet-stream';
			$filename  = "payment_records";
			header('Content-Encoding: UTF-8');
			header('Content-Type: ' . $mime_type . ' ;charset=UTF-8');
			header('Expires: ' . gmdate('D, d M Y H:i:s') . ' GMT');

			if ($UserBrowser == 'IE')
			{
				header('Content-Disposition: attachment; filename="' . $filename . '.csv"');
				header('Cache-Control: must-revalidate, post-check=0, pre-check=0');
				header('Pragma: public');
			}
			else
			{
				header('Content-Disposition: attachment; filename="' . $filename . '.csv"');
				header('Pragma: no-cache');
			}

			$fp = fopen('php://output', 'w');
			fwrite($fp, "\xEF\xBB\xBF");
			$delimiter = isset($config->csv_delimiter) ? $config->csv_delimiter : ',';
			$fields    = [];
			$fields[]  = Text::_('Form');

			if (count($rowFields))
			{
				foreach ($rowFields as $rowField)
				{
					$fields[] = Text::_($rowField->title);
				}
			}
			$fields[] = Text::_('Total Amount');
			$fields[] = Text::_('Discount Amount');
			$fields[] = Text::_('Gross Amount');
			$fields[] = Text::_('Payment Date');
			$fields[] = Text::_('Payment method');
			$fields[] = Text::_('Transaction ID');

			if ($config->activate_invoice_feature)
			{
				$fields[] = Text::_('Invoice Number');
			}

			fputcsv($fp, $fields, $delimiter);

			foreach ($rows as $r)
			{
				$fields   = [];
				$fields[] = $r->title;

				foreach ($rowFields as $rowField)
				{
					if ($rowField->is_core)
					{
						$fields[] = @$r->{$rowField->name};
					}
					else
					{
						$fieldValue = @$fieldValues[$r->id][$rowField->id];

						if (is_string($fieldValue) && is_array(json_decode($fieldValue)))
						{
							$fieldValue = implode(', ', json_decode($fieldValue));
						}
						$fields[] = $fieldValue;
					}
				}

				$fields[] = static::formatAmount($r->total_amount, $config);
				$fields[] = static::formatAmount($r->discount_amount, $config);
				$fields[] = static::formatAmount($r->gross_amount, $config);
				$fields[] = HTMLHelper::_('date', $r->created_date, $config->date_format, null);
				$method   = PaymentMethods::getPaymentMethod($r->payment_method);

				if ($method)
				{
					$fields[] = $method->getTitle();
				}
				else
				{
					$fields[] = '';
				}

				$fields[] = $r->transaction_id;

				if ($config->activate_invoice_feature)
				{
					if ($r->invoice_number > 0)
					{
						$fields[] = static::formatInvoiceNumber($r->invoice_number, $config);
					}
					else
					{
						$fields[] = '';
					}
				}

				fputcsv($fp, $fields, $delimiter);
			}

			fclose($fp);

			Factory::getApplication()->close();
		}
	}

	/**
	 * Check if the current file is just used to allow users to enter amount to make payment
	 *
	 * @param $rowFields
	 *
	 * @return bool
	 */
	public static function isEnterAmount($rowFields)
	{
		foreach ($rowFields as $rowField)
		{
			if ($rowField->name == 'amount')
			{
				return true;
			}
		}

		return false;
	}

	/**
	 * Method to render common layout used in different views in the extension
	 *
	 * @param   string  $layout
	 * @param   array   $data
	 *
	 * @return string
	 * @throws \Exception
	 */
	public static function loadCommonLayout($layout, $data = [])
	{
		$app       = Factory::getApplication();
		$themeFile = str_replace('/tmpl', '', $layout);

		if (file_exists($layout))
		{
			$path = $layout;
		}
		elseif (file_exists(JPATH_THEMES . '/' . $app->getTemplate() . '/html/com_pmform/' . $themeFile))
		{
			$path = JPATH_THEMES . '/' . $app->getTemplate() . '/html/com_pmform/' . $themeFile;
		}
		elseif (file_exists(JPATH_ROOT . '/components/com_pmform/View/' . $layout))
		{
			$path = JPATH_ROOT . '/components/com_pmform/View/' . $layout;
		}
		else
		{
			throw new \RuntimeException(Text::_('The given shared template path is not exist'));
		}

		// Start an output buffer.
		ob_start();
		extract($data);
		// Load the layout.
		include $path;

		// Get the layout contents.
		return ob_get_clean();
	}

	/**
	 * Check to see whether we need to generate a invoice for a payment record
	 *
	 * @param $row
	 *
	 * @return bool
	 */
	public static function needInvoice($row)
	{
		if ($row->amount > 0 || $row->gross_amount > 0)
		{
			return true;
		}

		return false;
	}

	/**
	 * Get the invoice number for this subscription record
	 */
	public static function getInvoiceNumber()
	{
		$config = self::getConfig();
		$db     = Factory::getDbo();
		$query  = $db->getQuery(true)
			->select('MAX(invoice_number)')
			->from('#__pf_payments');
		$db->setQuery($query);
		$invoiceNumber = (int) $db->loadResult();
		$invoiceNumber++;

		return max($invoiceNumber, (int) $config->invoice_start_number);
	}

	/**
	 * Generate invoice PDF (receipt) for a donation record
	 *
	 * @param $row
	 *
	 * @return array
	 */
	public static function generateInvoicePDF($row)
	{
		self::loadLanguage();
		$config   = self::getConfig();
		$siteName = Factory::getApplication()->get("sitename");

		require_once JPATH_ROOT . '/components/com_pmform/tcpdf/config/tcpdf_config.php';
		\JLoader::register('TCPDF', JPATH_ROOT . '/components/com_pmform/tcpdf/tcpdf.php');

		$pdf = new \TCPDF(PDF_PAGE_ORIENTATION, PDF_UNIT, PDF_PAGE_FORMAT, true, 'UTF-8', false);
		$pdf->setCreator(PDF_CREATOR);
		$pdf->setAuthor($siteName);
		$pdf->setTitle('Invoice');
		$pdf->setSubject('Invoice');
		$pdf->setKeywords('Invoice');
		$pdf->setHeaderFont([PDF_FONT_NAME_MAIN, '', PDF_FONT_SIZE_MAIN]);
		$pdf->setFooterFont([PDF_FONT_NAME_DATA, '', PDF_FONT_SIZE_DATA]);
		$pdf->setPrintHeader(false);
		$pdf->setPrintFooter(false);
		$pdf->SetMargins(PDF_MARGIN_LEFT, 0, PDF_MARGIN_RIGHT);
		$pdf->SetHeaderMargin(PDF_MARGIN_HEADER);
		$pdf->SetFooterMargin(PDF_MARGIN_FOOTER);
		//set auto page breaks
		$pdf->SetAutoPageBreak(true, PDF_MARGIN_BOTTOM);
		//set image scale factor
		$pdf->setImageScale(PDF_IMAGE_SCALE_RATIO);

		$font = empty($config->pdf_font) ? 'times' : $config->pdf_font;

		// True type font
		if (substr($font, -4) == '.ttf')
		{
			$font = \TCPDF_FONTS::addTTFfont(JPATH_ROOT . '/components/com_pmform/tcpdf/fonts/' . $font, 'TrueTypeUnicode', '', 96);
		}

		$pdf->SetFont($font, '', 8);
		$pdf->AddPage();

		$fieldSuffix = static::getFieldSuffix($row->language);

		static::prepareMultilingualMessages($config, ['invoice_format'], $fieldSuffix);

		$invoiceOutput = $config->invoice_format;

		$replaces = self::callOverridableHelperMethod('Helper', 'buildReplaceTags', [$row, $config]);

		unset($replaces['amount'], $replaces['total_amount'], $replaces['tax_amount']);

		$itemName                    = Text::_('PF_PAYMENT_FOR');
		$itemName                    = str_replace('[FORM_TITLE]', $replaces['form_title'], $itemName);
		$replaces['ITEM_NAME']       = $itemName;
		$replaces['NAME']            = $row->first_name . ' ' . $row->last_name;
		$replaces['ITEM_QUANTITY']   = 1;
		$replaces['AMOUNT']          = $replaces['ITEM_AMOUNT'] = $replaces['ITEM_SUB_TOTAL'] = self::formatAmount($row->total_amount, $config);
		$replaces['DISCOUNT_AMOUNT'] = self::formatAmount($row->discount_amount, $config);
		$replaces['SUB_TOTAL']       = self::formatAmount($row->total_amount - $row->discount_amount, $config);
		$replaces['TOTAL_AMOUNT']    = self::formatAmount($row->gross_amount, $config);
		$replaces['TAX_AMOUNT']      = self::formatAmount($row->tax_amount, $config);

		// Invoice Status
		if ($row->published == 0)
		{
			$invoiceStatus = Text::_('PF_INVOICE_STATUS_PENDING');
		}
		elseif ($row->published == 1)
		{
			$invoiceStatus = Text::_('PF_INVOICE_STATUS_PAID');
		}
		else
		{
			$invoiceStatus = Text::_('PF_INVOICE_STATUS_UNKNOWN');
		}
		$replaces['INVOICE_STATUS'] = $invoiceStatus;
		$replaces['invoice_date']   = HTMLHelper::_('date', $row->created_date, $config->date_format);

		foreach ($replaces as $key => $value)
		{
			$key           = strtoupper($key);
			$invoiceOutput = str_replace("[$key]", $value, $invoiceOutput);
		}

		$invoiceOutput = self::convertImgTags($invoiceOutput);
		$pdf->writeHTML($invoiceOutput, true, false, false, false, '');
		//Filename
		$filePath = JPATH_ROOT . '/media/com_pmform/invoices/' . $replaces['invoice_number'] . '.pdf';
		$pdf->Output($filePath, 'F');

		return [$filePath, $replaces['invoice_number'] . '.pdf'];
	}

	/**
	 * Format invoice number according to configuration
	 *
	 * @param $invoiceNumber
	 * @param $config
	 *
	 * @return string
	 */
	public static function formatInvoiceNumber($invoiceNumber, $config)
	{
		return $config->invoice_prefix . str_pad($invoiceNumber, $config->invoice_number_length ? $config->invoice_number_length : 4, '0', STR_PAD_LEFT);
	}

	/**
	 * Convert src of img tags to use absolute links instead of ralative link
	 *
	 * @param $html_content
	 *
	 * @return mixed
	 */
	public static function convertImgTags($html_content)
	{
		$patterns     = [];
		$replacements = [];
		$i            = 0;
		$src_exp      = "/src=\"(.*?)\"/";
		$link_exp     = "[^http:\/\/www\.|^www\.|^https:\/\/|^http:\/\/]";
		$siteURL      = Uri::root();

		preg_match_all($src_exp, $html_content, $out, PREG_SET_ORDER);

		foreach ($out as $val)
		{
			$links = preg_match($link_exp, $val[1], $match, PREG_OFFSET_CAPTURE);

			if ($links == '0')
			{
				$patterns[$i]     = $val[1];
				$patterns[$i]     = "\"$val[1]";
				$replacements[$i] = $siteURL . $val[1];
				$replacements[$i] = "\"$replacements[$i]";
			}

			$i++;
		}

		return str_replace($patterns, $replacements, $html_content);
	}

	public static function loadTooltip($selector, $params)
	{
		$opt['animation'] = (isset($params['animation']) && ($params['animation'])) ? (boolean) $params['animation'] : null;
		$opt['html']      = (isset($params['html']) && ($params['html'])) ? (boolean) $params['html'] : null;
		$opt['placement'] = (isset($params['placement']) && ($params['placement'])) ? (string) $params['placement'] : null;
		$opt['selector']  = (isset($params['selector']) && ($params['selector'])) ? (string) $params['selector'] : null;
		$opt['title']     = (isset($params['title']) && ($params['title'])) ? (string) $params['title'] : null;
		$opt['trigger']   = (isset($params['trigger']) && ($params['trigger'])) ? (string) $params['trigger'] : null;
		$opt['delay']     = (isset($params['delay']) && ($params['delay'])) ? (int) $params['delay'] : null;

		$options = self::getJSObject($opt);

		// Attach tooltips to document
		Factory::getDocument()->addScriptDeclaration(
			"jQuery(document).ready(function() {
			jQuery('" . $selector . "').tooltip(" . $options . ");
		});"
		);
	}


	/**
	 * Load jquery used in the extension
	 *
	 */
	public static function loadJQuery()
	{
		HTMLHelper::_('jquery.framework');
	}

	/**
	 * Load twitter bootstrap framework
	 *
	 * @param   bool  $loadJs
	 */
	public static function loadBootstrap($loadJs = true)
	{
		$document = Factory::getDocument();

		if ($loadJs)
		{
			self::loadJQuery();

			$document->addScript(Uri::root(true) . '/media/com_pmform/assets/bootstrap/js/bootstrap.min.js');
		}

		$document->addStyleSheet(Uri::root(true) . '/media/com_pmform/assets/bootstrap/css/bootstrap.min.css');
	}

	/**
	 * Create and initialize mailer object from configuration data
	 *
	 * @return \JMail
	 */
	public static function getMailer()
	{
		$mailer = Factory::getMailer();
		$config = static::getConfig();

		if ($config->from_name)
		{
			$fromName = $config->from_name;
		}
		else
		{
			$fromName = Factory::getApplication()->get('fromname');
		}

		if ($config->from_email)
		{
			$fromEmail = $config->from_email;
		}
		else
		{
			$fromEmail = Factory::getApplication()->get('mailfrom');
		}

		$mailer->setSender([$fromEmail, $fromName]);

		$mailer->isHtml(true);

		if (!$config->notification_emails)
		{
			$config->notification_emails = $fromEmail;
		}

		return $mailer;
	}

	/**
	 * Process sending after all the data has been initialized
	 *
	 * @param   \JMail  $mailer
	 * @param   array   $emails
	 * @param   string  $subject
	 * @param   string  $body
	 *
	 */
	public static function send($mailer, $emails, $subject, $body)
	{
		if (empty($subject))
		{
			return;
		}

		$emails = (array) $emails;

		$emails = array_map('trim', $emails);

		for ($i = 0, $n = count($emails); $i < $n; $i++)
		{
			if (!MailHelper::isEmailAddress($emails[$i]))
			{
				unset($emails[$i]);
			}
		}

		$emails = array_unique($emails);

		if (count($emails) == 0)
		{
			return;
		}

		$email = $emails[0];
		$mailer->addRecipient($email);

		if (count($emails) > 1)
		{
			unset($emails[0]);
			$bccEmails = $emails;
			$mailer->addBcc($bccEmails);
		}

		$body = static::processConditionalText($body);

		try
		{
			$mailer->setSubject($subject)
				->setBody($body)
				->Send();
		}
		catch (\Exception $e)
		{
			// Ignore for now
		}
	}

	/**
	 * Basic implement for conditional text
	 *
	 * @param   string  $text
	 *
	 * @return string mixed
	 */
	public static function processConditionalText($text)
	{
		$regex = '#{pmfShowText (.+)(=|>|<)(.+)}(.+){/pmfShowText}#';

		return preg_replace_callback($regex, 'static::processCondition', $text);
	}

	/**
	 * Process conditional text, for now, we support = , >, and < operator
	 *
	 * @param   array  $matches
	 *
	 * @return string
	 */
	public static function processCondition($matches)
	{
		$a        = trim($matches[1]);
		$operator = $matches[2];
		$b        = trim($matches[3]);
		$text     = $matches[4];

		switch ($operator)
		{
			case '=':
				if ($a == $b) return $text;
				break;
			case '>':
				if ($a > $b) return $text;
				break;
			case '<':
				if ($a < $b) return $text;
				break;
				break;
		}

		return null;
	}

	public static function getJSObject(array $array = [])
	{
		$object = '{';

		// Iterate over array to build objects
		foreach ((array) $array as $k => $v)
		{
			if (is_null($v))
			{
				continue;
			}

			if (is_bool($v))
			{
				$object .= ' ' . $k . ': ';
				$object .= ($v) ? 'true' : 'false';
				$object .= ',';
			}
			elseif (!is_array($v) && !is_object($v))
			{
				$object .= ' ' . $k . ': ';
				$object .= (is_numeric($v) || strpos($v, '\\') === 0) ? (is_numeric($v)) ? $v : substr($v, 1) : "'" . $v . "'";
				$object .= ',';
			}
			else
			{
				$object .= ' ' . $k . ': ' . self::getJSObject($v) . ',';
			}
		}

		if (substr($object, -1) == ',')
		{
			$object = substr($object, 0, -1);
		}

		$object .= '}';

		return $object;
	}

	/**
	 *
	 *
	 * @return string
	 */
	public static function validateEngine()
	{
		$config     = self::getConfig();
		$dateFormat = $config->date_field_format ? $config->date_field_format : '%Y-%m-%d';
		$dateFormat = str_replace('%', '', $dateFormat);
		$dateNow    = HTMLHelper::_('date', Factory::getDate(), $dateFormat);

		$validClass = [
			"",
			"validate[custom[integer]]",
			"validate[custom[number]]",
			"validate[custom[email]]",
			"validate[custom[url]]",
			"validate[custom[phone]]",
			"validate[custom[date],past[$dateNow]]",
			"validate[custom[ipv4]]",
			"validate[minSize[6]]",
			"validate[maxSize[12]]",
			"validate[custom[integer],min[-5]]",
			"validate[custom[integer],max[50]]"];

		return json_encode($validClass);
	}

	/**
	 *
	 *
	 * @return string
	 */
	public static function getValidationRules()
	{
		$config     = self::getConfig();
		$dateFormat = $config->date_field_format ? $config->date_field_format : '%Y-%m-%d';
		$dateFormat = str_replace('%', '', $dateFormat);
		$dateNow    = HTMLHelper::_('date', 'now', $dateFormat);

		return [
			"",
			"custom[integer]",
			"custom[number]",
			"custom[email]",
			"custom[url]",
			"custom[phone]",
			"custom[date],past[$dateNow]",
			"custom[ipv4]",
			"minSize[6]",
			"maxSize[12]",
			"custom[integer],min[-5]",
			"custom[integer],max[50]",
		];
	}

	public static function getUserInput($userId, $fieldName = 'user_id')
	{
		// Initialize variables.
		$html = [];
		$link = 'index.php?option=com_users&amp;view=users&amp;layout=modal&amp;tmpl=component&amp;field=user_id';
		// Initialize some field attributes.
		$attr = ' class="form-control"';
		// Load the modal behavior script.
		HTMLHelper::_('behavior.modal', 'a.modal_user_id');
		// Build the script.
		$script   = [];
		$script[] = '	function jSelectUser_user_id(id, title) {';
		$script[] = '		var old_id = document.getElementById("user_id").value;';
		$script[] = '		if (old_id != id) {';
		$script[] = '			document.getElementById("' . $fieldName . '").value = id;';
		$script[] = '			document.getElementById("user_id_name").value = title;';
		$script[] = '		}';
		$script[] = '		SqueezeBox.close();';
		$script[] = '	}';
		// Add the script to the document head.
		Factory::getDocument()->addScriptDeclaration(implode("\n", $script));

		// Load the current username if available.
		$table = Table::getInstance('user');

		if ($userId)
		{
			$table->load($userId);
		}
		else
		{
			$table->name = '';
		}

		// Create a dummy text field with the user name.
		$html[] = '<div class="fltlft">';
		$html[] = '	<input type="text" id="user_id_name"' . ' value="' . htmlspecialchars($table->name, ENT_COMPAT, 'UTF-8') . '"' .
			' disabled="disabled"' . $attr . ' />';
		$html[] = '</div>';
		// Create the user select button.
		$html[] = '<div class="button2-left">';
		$html[] = '<div class="blank">';
		$html[] = '<a class="modal_user_id" title="' . Text::_('JLIB_FORM_CHANGE_USER') . '"' . ' href="' . $link . '"' .
			' rel="{handler: \'iframe\', size: {x: 800, y: 500}}">';
		$html[] = '	' . Text::_('JLIB_FORM_CHANGE_USER') . '</a>';
		$html[] = '</div>';
		$html[] = '</div>';
		// Create the real field, hidden, that stored the user id.
		$html[] = '<input type="hidden" id="' . $fieldName . '" name="' . $fieldName . '" value="' . $userId . '" />';

		return implode("\n", $html);
	}

	/**
	 * Render showon string
	 *
	 * @param   array  $fields
	 *
	 * @return string
	 */
	public static function renderShowon($fields)
	{
		$output = [];

		$i = 0;

		foreach ($fields as $name => $values)
		{
			$i++;

			$values = (array) $values;

			$data = [
				'field'  => $name,
				'values' => $values,
			];

			if (version_compare(JVERSION, '3.6.99', 'ge'))
			{
				$data['sign'] = '=';
			}

			$data['op'] = $i > 1 ? 'AND' : '';

			$output[] = json_encode($data);
		}

		return '[' . implode(',', $output) . ']';
	}

	/**
	 * Generate article selection box
	 *
	 * @param   int     $fieldValue
	 * @param   string  $fieldName
	 *
	 * @return string
	 */
	public static function getArticleInput($fieldValue, $fieldName = 'article_id')
	{
		HTMLHelper::_('jquery.framework');
		JFormHelper::addFieldPath(JPATH_ADMINISTRATOR . '/components/com_content/models/fields');

		if (static::isJoomla4())
		{
			JFormHelper::addFieldPrefix('Joomla\Component\Content\Administrator\Field');

		}

		$field = JFormHelper::loadFieldType('Modal_Article');

		$element = new \SimpleXMLElement('<field />');
		$element->addAttribute('name', $fieldName);
		$element->addAttribute('select', 'true');
		$element->addAttribute('clear', 'true');

		$field->setup($element, $fieldValue);

		return $field->input;
	}

	/**
	 * Function to find Itemid
	 *
	 * @param   array  $needles
	 * @param   int    $itemId
	 *
	 * @return int
	 */
	public static function findItem($needles = [], $itemId = 0)
	{
		static $lookup;

		// Prepare the reverse lookup array.
		if ($lookup === null)
		{
			$lookup    = [];
			$component = ComponentHelper::getComponent('com_pmform');
			$menus     = Factory::getApplication()->getMenu('site');
			$items     = $menus->getItems('component_id', $component->id);

			foreach ($items as $item)
			{
				if (empty($item->query['view']))
				{
					continue;
				}

				$view = $item->query['view'];

				if (!isset($lookup[$view]))
				{
					$lookup[$view] = [];
				}

				if (isset($item->query['form_id']))
				{
					$lookup[$view][$item->query['form_id']] = $item->id;
				}
				else
				{
					$lookup[$view][0] = $item->id;
				}
			}
		}

		if ($needles)
		{
			foreach ($needles as $view => $ids)
			{
				if (isset($lookup[$view]))
				{
					foreach ($ids as $id)
					{
						$id = (int) $id;

						if (isset($lookup[$view][$id]))
						{
							return $lookup[$view][$id];
						}
					}
				}
			}
		}

		//Return default item id
		return $itemId;
	}

	/**
	 * Method to encrypt a string
	 *
	 * @param   string  $string
	 *
	 * @return string
	 */
	public static function encrypt($string)
	{
		$privateKey = md5(Factory::getApplication()->get('secret'));

		if (class_exists('\JCryptCipherSimple'))
		{
			$key   = new \JCryptKey('simple', $privateKey, $privateKey);
			$crypt = new \JCrypt(new \JCryptCipherSimple, $key);

			return $crypt->encrypt($string);
		}

		return $string;
	}

	/**
	 * Method to decrypt a string
	 *
	 * @param $string
	 *
	 * @return string
	 */
	public static function decrypt($string)
	{
		$privateKey = md5(Factory::getConfig()->get('secret'));

		if (class_exists('\JCryptCipherSimple'))
		{
			$key   = new \JCryptKey('simple', $privateKey, $privateKey);
			$crypt = new \JCrypt(new \JCryptCipherSimple, $key);

			return $crypt->decrypt($string);
		}

		return $string;
	}

	/**
	 * Add choices JS to dropdown
	 *
	 * @param   string  $html
	 * @param   string  $hint
	 */
	public static function getChoicesJsSelect($html, $hint = '')
	{
		static $isJoomla4;

		if ($isJoomla4 === null)
		{
			$isJoomla4 = self::isJoomla4();
		}

		if ($isJoomla4)
		{
			Text::script('JGLOBAL_SELECT_NO_RESULTS_MATCH');
			Text::script('JGLOBAL_SELECT_PRESS_TO_SELECT');

			Factory::getApplication()->getDocument()->getWebAssetManager()
				->usePreset('choicesjs')
				->useScript('webcomponent.field-fancy-select');

			$attributes = [];

			$hint = $hint ?: Text::_('JGLOBAL_TYPE_OR_SELECT_SOME_OPTIONS');

			$attributes[] = 'placeholder="' . $hint . '""';
			$attributes[] = 'search-placeholder="' . $hint . '""';


			return '<joomla-field-fancy-select ' . implode(' ', $attributes) . '>' . $html . '</joomla-field-fancy-select>';
		}

		return $html;
	}
}