<?php
/**
 * @version        6.2.0
 * @package        Joomla
 * @subpackage     Payment Form
 * @author         Tuan Pham Ngoc
 * @copyright      Copyright (C) 2010 - 2021 Ossolution Team
 * @license        GNU/GPL, see LICENSE.php
 */

namespace OSSolution\PaymentForm\Site\Helper;

use Joomla\CMS\Factory;
use Joomla\CMS\Filesystem\File;
use Joomla\CMS\Uri\Uri;
use Joomla\Registry\Registry;

defined('_JEXEC') or die;

class PaymentMethods
{
	/**
	 * Get list of payment methods
	 *
	 * @return array
	 */
	public static function getPaymentMethods($paymentMethods = null)
	{
		static $methods;

		if (!$methods)
		{
			$db    = Factory::getDbo();
			$query = $db->getQuery(true);
			$query->select('*')
				->from('#__pf_payment_plugins')
				->where('published = 1');

			if ($paymentMethods)
			{
				$query->where('id IN (' . $paymentMethods . ')');
			}

			$query->order('ordering');
			$db->setQuery($query);
			$rows = $db->loadObjectList();

			$baseUri = Uri::base(true);

			foreach ($rows as $row)
			{
				$filePath = JPATH_ROOT . '/components/com_pmform/payments/' . $row->name . '.php';

				if (!File::exists($filePath))
				{
					throw new \RuntimeException(printf('Payment Class for payment method %s does not exist', $row->title));
				}

				require_once $filePath;

				$params = new Registry($row->params);

				/* @var \POSFPayment $method */
				$method = new $row->name($params);
				$method->setTitle($row->title);
				$method->paymentFee = false;

				if ($params->get('payment_fee_amount') > 0 || $params->get('payment_fee_percent'))
				{
					$method->paymentFee = true;
				}

				$iconUri = '';

				if ($icon = $params->get('icon'))
				{
					if (file_exists(JPATH_ROOT . '/media/com_pmform/assets/paymentmethods/' . $icon))
					{
						$iconUri = $baseUri . '/media/com_pmform/assets/paymentmethods/' . $icon;
					}
					elseif (file_exists(JPATH_ROOT . '/' . $icon))
					{
						$iconUri = $baseUri . '/' . $icon;
					}
				}

				$method->iconUri = $iconUri;

				$methods[] = $method;
			}
		}

		return $methods;
	}

	/**
	 * Write the javascript objects to show the page
	 *
	 * @return string
	 */
	public static function writeJavascriptObjects()
	{
		$methods = self::getPaymentMethods();

		$jsString = " methods = new PaymentMethods();\n";

		if (count($methods))
		{
			foreach ($methods as $method)
			{
				$jsString .= " method = new PaymentMethod('" . $method->getName() . "'," . $method->getCreditCard() . "," . $method->getCardType() . "," . $method->getCardCvv() . "," . $method->getCardHolderName() . ");\n";
				$jsString .= " methods.Add(method);\n";
			}
		}

		Factory::getDocument()->addScriptDeclaration($jsString);
	}

	/**
	 * Load information about the payment method
	 *
	 * @param   string  $name
	 *
	 * @return \OSSolution\PaymentForm\Admin\Table\Plugin
	 */
	public static function loadPaymentMethod($name)
	{
		$db    = Factory::getDbo();
		$query = $db->getQuery(true);
		$query->select('*')
			->from('#__pf_payment_plugins')
			->where('name = ' . $db->quote($name));
		$db->setQuery($query);

		return $db->loadObject();
	}

	/**
	 * Get default payment method. It is the first published payment method in the extension
	 *
	 * @return string
	 */
	public static function getDefaultPaymentMethod($paymentMethods = null)
	{
		$db    = Factory::getDbo();
		$query = $db->getQuery(true);
		$query->select('name')
			->from('#__pf_payment_plugins')
			->where('published = 1');

		if ($paymentMethods)
		{
			$query->where('id IN (' . $paymentMethods . ')');
		}

		$query->order('ordering');
		$db->setQuery($query);

		return $db->loadResult();
	}

	/**
	 * Get the payment method object based on it's name
	 *
	 * @param   string  $name
	 *
	 * @return \POSFPayment
	 */
	public static function getPaymentMethod($name)
	{
		$methods = self::getPaymentMethods();

		foreach ($methods as $method)
		{
			if ($method->getName() == $name)
			{
				return $method;
			}
		}

		return null;
	}
}