<?php
/**
 * @version        6.2.0
 * @package        Joomla
 * @subpackage     Payment Form
 * @author         Tuan Pham Ngoc
 * @copyright      Copyright (C) 2010 - 2021 Ossolution Team
 * @license        GNU/GPL, see LICENSE.php
 */

namespace OSSolution\PaymentForm\Site\Model;

use Joomla\CMS\Factory;
use Joomla\CMS\HTML\HTMLHelper;
use Joomla\CMS\Language\Multilanguage;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\PluginHelper;
use Joomla\CMS\Router\Route;
use Joomla\CMS\User\UserHelper;
use Joomla\Registry\Registry;
use OSL\Model\Model;
use OSSolution\PaymentForm\Site\Helper\Helper as PMFormHelper;

defined('_JEXEC') or die;

class Payment extends Model
{
	/**
	 * Process payment
	 *
	 * @param   \OSL\Input\Input  $input
	 */
	public function processPayment($input)
	{
		jimport('joomla.user.helper');
		$user                   = $this->container->user;
		$db                     = $this->getDbo();
		$query                  = $db->getQuery(true);
		$config                 = PMFormHelper::getConfig();
		$userId                 = $user->id;
		$data                   = $input->getData();
		$data['transaction_id'] = strtoupper(UserHelper::genRandomPassword());
		$data['order_number']   = strtoupper(UserHelper::genRandomPassword(12));
		$row                    = $this->getTable('Payment');
		$row->bind($data);
		$row->ip_address   = $input->server->get('REMOTE_ADDR');
		$row->published    = 0;
		$row->created_date = gmdate('Y-m-d H:i:s');

		$formId      = (int) $row->form_id;
		$currentDate = HTMLHelper::_('date', 'now', 'Y-m-d');
		$query->select('*, DATEDIFF("' . $currentDate . '", late_payment_date) AS number_days')
			->from('#__pf_forms')
			->where('id = ' . $formId);
		$db->setQuery($query);
		$rowForm = $db->loadObject();

		if ((int) $rowForm->user_registration !== 2)
		{
			$config->user_registration = $rowForm->user_registration;
		}

		if (!$userId && $config->user_registration)
		{
			if ($config->create_account_when_payment_success !== '1')
			{
				$userId = PMFormHelper::saveRegistration($data);
			}
			else
			{
				$row->user_password = PMFormHelper::encrypt($data['password1']);
			}
		}

		$row->user_id = $userId;


		if (empty($rowForm) || !$rowForm->published)
		{
			return;
		}

		// Check access level of the form
		if (!in_array($rowForm->access, $user->getAuthorisedViewLevels()))
		{
			throw new \Exception('You do not have permission to process payment on this form', 403);
		}

		$rowFields = PMFormHelper::getFormFields($formId);
		$form      = new \POSFForm($rowFields);
		$form->bind($data);
		$paymentMethod = isset($data['payment_method']) ? $data['payment_method'] : '';
		$fees          = PMFormHelper::callOverridableHelperMethod('Helper', 'calculateFormFee', [$rowForm, $form, $input, $paymentMethod]);

		$row->total_amount           = $fees['total_amount'];
		$row->discount_amount        = $fees['discount_amount'];
		$row->tax_amount             = $fees['tax_amount'];
		$row->late_fee               = $fees['late_fee'];
		$row->gross_amount           = $fees['gross_amount'];
		$row->payment_processing_fee = $fees['payment_processing_fee'];
		$row->coupon_id              = (int) $fees['coupon_id'];

		//Save the active language
		if (Multilanguage::isEnabled())
		{
			$row->language = Factory::getLanguage()->getTag();
		}
		else
		{
			$row->language = '*';
		}

		$row->store();

		$this->container->session->set('pmf_payment_id', $row->id);

		//Store custom fields data
		PMFormHelper::storeFormData($row, $data);
		$couponId = (int) $fees['coupon_id'];

		if ($couponId)
		{
			//Update number of times the coupon has used
			$sql = 'UPDATE #__pf_coupons SET used = used +1 WHERE id=' . $couponId;
			$db->setQuery($sql);
			$db->execute();
		}

		PluginHelper::importPlugin('pmform');
		$this->container->app->triggerEvent('onAfterStorePayment', [$row]);

		if ($row->gross_amount > 0)
		{
			$data['amount'] = $row->gross_amount;

			if ($rowForm->paypal_account)
			{
				$data['business'] = $rowForm->paypal_account;
			}

			$paymentMethod = $data['payment_method'];
			$itemName      = Text::_('PF_PAYMENT_FOR');
			$itemName      = str_replace('[FORM_TITLE]', $rowForm->title, $itemName);
			$itemName      = str_replace('[FIRST_NAME]', $row->first_name, $itemName);
			$itemName      = str_replace('[LAST_NAME]', $row->last_name, $itemName);
			$itemName      = str_replace('[PAYMENT_ID]', $row->id, $itemName);

			$replaces = PMFormHelper::callOverridableHelperMethod('Helper', 'buildReplaceTags', [$row, $config]);

			foreach ($replaces as $key => $value)
			{
				$key      = strtoupper($key);
				$itemName = str_replace("[$key]", $value, $itemName);
			}

			$data['item_name'] = $itemName;

			require_once JPATH_COMPONENT . '/payments/' . $paymentMethod . '.php';

			// Guess card type based on card number
			if (!empty($data['x_card_num']) && empty($data['card_type']))
			{
				$data['card_type'] = \OSL\Utils\Helper::getCardType($data['x_card_num']);
			}

			$query->clear()
				->select('title, params')
				->from('#__pf_payment_plugins')
				->where('name = ' . $db->quote($paymentMethod));
			$db->setQuery($query);
			$rowPaymentMethod = $db->loadObject();
			$params           = new Registry($rowPaymentMethod->params);

			/* @var \POSFPayment $paymentClass */
			$paymentClass = new $paymentMethod($params);
			$paymentClass->setTitle($rowPaymentMethod->title);

			if ($rowForm->currency_code)
			{
				$currency = $rowForm->currency_code;
			}
			else
			{
				$currency = $config->currency_code;
			}

			$data['amount']   = round($data['amount'], 2);
			$data['currency'] = $currency;
			$country          = empty($data['country']) ? $config->default_country : $data['country'];
			$data['country']  = PMFormHelper::getCountryCode($country);

			// Store payment amount and payment currency for PayPal payment verification
			$row->payment_amount   = $data['amount'];
			$row->payment_currency = $currency;
			$row->store();

			$paymentClass->processPayment($row, $data);
		}
		else
		{
			$Itemid            = $input->getInt('Itemid', 0);
			$row->payment_date = gmdate('Y-m-d H:i:s');
			$row->published    = 1;
			$row->store();

			$this->container->app->triggerEvent('onAfterPaymentSuccess', [$row]);

			PMFormHelper::sendEmails($row, $config);

			$url = Route::_('index.php?option=com_pmform&view=complete&id=' . $row->id . '&Itemid=' . $Itemid, false);
			$this->container->app->redirect($url);
		}
	}
}
