<?php
/**
 * @version        6.2.0
 * @package        Joomla
 * @subpackage     Payment Form
 * @author         Tuan Pham Ngoc
 * @copyright      Copyright (C) 2010 - 2021 Ossolution Team
 * @license        GNU/GPL, see LICENSE.php
 */

namespace OSSolution\PaymentForm\Site\View\Confirmation;

use Joomla\CMS\Filesystem\File;
use Joomla\CMS\Filesystem\Folder;
use Joomla\CMS\HTML\HTMLHelper;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Router\Route;
use OSL\Utils\Database as DatabaseUtils;
use OSL\View\HtmlView;
use OSSolution\PaymentForm\Site\Helper\Helper as PMFormHelper;
use OSSolution\PaymentForm\Site\Helper\PaymentMethods as PaymentMethodsHelper;
use PMFormHelperBootstrap;

defined('_JEXEC') or die;

class Html extends HtmlView
{
	/**
	 * Tell the controller that we don't need a model, so no need to create it
	 *
	 * @var bool
	 */
	public $hasModel = false;

	/**
	 * Display the form
	 */

	public function display()
	{
		$app         = $this->container->app;
		$user        = $this->container->user;
		$document    = $this->container->document;
		$input       = $this->input;
		$db          = $this->container->db;
		$query       = $db->getQuery(true);
		$config      = PMFormHelper::getConfig();
		$userId      = $user->get('id');
		$formId      = $input->getInt('form_id', 0);
		$fieldSuffix = PMFormHelper::getFieldSuffix();

		if (!$formId)
		{
			$app->enqueueMessage(Text::_('PMF_INVALID_FORM'), 'error');
			$app->redirect(Route::_('index.php'), 404);
		}

		$currentDate = HTMLHelper::_('date', 'now', 'Y-m-d');
		$query->select('*, DATEDIFF("' . $currentDate . '", late_payment_date) AS number_days')
			->from('#__pf_forms')
			->where('id=' . $formId);

		if ($fieldSuffix)
		{
			DatabaseUtils::getMultilingualFields($query, ['title'], $fieldSuffix);
		}

		$db->setQuery($query);
		$rowForm = $db->loadObject();

		if (!$rowForm || $rowForm->published == 0)
		{
			$app->enqueueMessage(Text::_('PMF_INVALID_FORM'), 'error');
			$app->redirect(Route::_('index.php'), 404);
		}

		// Pass user registration setting from form to configuration
		if ((int) $rowForm->user_registration !== 2)
		{
			$config->user_registration = $rowForm->user_registration;
		}

		$rowFields = PMFormHelper::getFormFields($formId, $fieldSuffix);
		$data      = $input->post->getData();

		$uploadFolder = $config->path_upload;
		if (!$uploadFolder)
		{
			$uploadFolder = 'images/com_pmform';
		}

		$pathUpload = JPATH_ROOT . '/' . $uploadFolder;

		if (!Folder::exists($pathUpload))
		{
			Folder::create($pathUpload);
		}

		$allowedExtensions = $config->allowed_extensions;

		if (!$allowedExtensions)
		{
			$allowedExtensions = 'doc, docx, ppt, pptx, pdf, zip, rar, jpg, jepg, png, zipx';
		}

		$allowedExtensions = explode(',', $allowedExtensions);
		$allowedExtensions = array_map('trim', $allowedExtensions);

		// Process field upload for file custom field types
		foreach ($rowFields as $rowField)
		{
			if ($rowField->fieldtype == 'File')
			{
				$name = $rowField->name;

				// If there are field, we need to upload the file to server and save it !
				if (isset($_FILES[$name]))
				{
					if ($_FILES[$name]['name'] != '')
					{
						$fileName = $_FILES[$name]['name'];
						$fileExt  = File::getExt($fileName);

						if (in_array(strtolower($fileExt), $allowedExtensions))
						{
							$fileName = File::makeSafe($fileName);

							if (File::exists($pathUpload . '/' . $fileName))
							{
								$targetFileName = time() . '_' . $fileName;
							}
							else
							{
								$targetFileName = $fileName;
							}

							if (File::upload($_FILES[$name]['tmp_name'], $pathUpload . '/' . $targetFileName))
							{
								$data[$name] = $targetFileName;
							}
						}
					}
				}
			}
		}

		//Get data
		$form = new \POSFForm($rowFields);
		$form->bind($data, false);
		$form->buildFieldsDependency();

		$fields = $form->getFields();

		foreach ($fields as $field)
		{
			if (!$field->visible)
			{
				unset($fields[$field->name]);
			}
		}

		$form->setFields($fields);

		if ($rowForm->payment_methods && strpos('0', $rowForm->payment_methods) === false)
		{
			$paymentMethod = $input->getString('payment_method', PaymentMethodsHelper::getDefaultPaymentMethod($rowForm->payment_methods));
		}
		else
		{
			$paymentMethod = $input->getString('payment_method', PaymentMethodsHelper::getDefaultPaymentMethod());
		}

		if ($rowForm->confirmation_msg)
		{
			$config->confirmation_message = $rowForm->confirmation_msg;
		}

		$method      = PaymentMethodsHelper::getPaymentMethod($paymentMethod);
		$fees = PMFormHelper::callOverridableHelperMethod('Helper', 'calculateFormFee', [$rowForm, $form, $input, $paymentMethod]);
		$hiddenInput = $form->renderHidden();

		// Handle output data on confirmation page
		$fields = $form->getFields();

		foreach ($fields as $field)
		{
			/* @var \POSFFormField $field */
			if (in_array($field->type, ['Checkboxes', 'Radio', 'List']))
			{
				$fieldValue = $field->value;

				if (is_array($fieldValue))
				{
					$fieldValue = json_encode($fieldValue);
				}

				$rowField               = new \stdClass;
				$rowField->values       = $field->row->values;
				$rowField->options_text = $field->row->options_text;

				$fieldValue = PMFormHelper::getTextValue($rowField, $fieldValue);

				$field->setValue($fieldValue);
			}
		}

		if (PMFormHelper::isEnterAmount($rowFields) && $fees['gross_amount'] == $fees['total_amount'])
		{
			$form->removeField('amount');
		}

		$pageTitle = Text::_('PF_PAYMENT_CONFIRMATION_PAGE_TITLE');
		$pageTitle = str_replace('[FORM_TITLE]', $rowForm->title, $pageTitle);
		$document->setTitle($pageTitle);

		if ($fieldSuffix)
		{
			PMFormHelper::prepareMultilingualMessages($rowForm, ['confirmation_msg'], $fieldSuffix);
			PMFormHelper::prepareMultilingualMessages($config, ['confirmation_message'], $fieldSuffix);
		}

		if (PMFormHelper::isValidMessage($rowForm->confirmation_msg))
		{
			$config->confirmation_message = $rowForm->confirmation_msg;
		}

		// Override currency symbol from form
		if ($rowForm->currency_symbol)
		{
			$config->currency_symbol = $rowForm->currency_symbol;
		}

		// Payment information
		$bankId         = $input->get('bank_id', '', 'none');
		$x_card_num     = $input->get('x_card_num', '', 'none');
		$expMonth       = $input->get('exp_month', '', 'none');
		$expYear        = $input->get('exp_year', '', 'none');
		$x_card_code    = $input->get('x_card_code', '', 'none');
		$cardHolderName = $input->getString('card_holder_name', '');
		$cardType       = $input->get('card_type', '', 'none');
		$x_exp_date     = str_pad($expMonth, 2, '0', STR_PAD_LEFT) . '/' . substr($expYear, 2, 2);

		$formPageUrl       = $input->get('form_page_url', '', 'none');
		$this->couponCode  = $input->getString('coupon_code', '');
		$this->form        = $form;
		$this->enterAmount = PMFormHelper::isEnterAmount($rowFields);

		// Assign these parameters
		$this->userId        = $userId;
		$this->paymentMethod = $paymentMethod;
		$this->config        = $config;
		$this->method        = $method;
		$this->formPageUrl   = $formPageUrl;
		$this->rowForm       = $rowForm;
		$this->fees          = $fees;

		$this->x_card_num     = $x_card_num;
		$this->x_card_code    = $x_card_code;
		$this->x_exp_date     = $x_exp_date;
		$this->expMonth       = $expMonth;
		$this->expYear        = $expYear;
		$this->cardHolderName = $cardHolderName;
		$this->stripeToken    = $input->getString('stripeToken', '');

		$this->bankId          = $bankId;
		$this->cardType        = $cardType;
		$this->username        = $input->get('username', '', 'nome');
		$this->password        = $input->get('password1', '', 'nome');
		$this->hiddenInput     = $hiddenInput;
		$this->bootstrapHelper = PMFormHelperBootstrap::getInstance();

		parent::display();
	}
}