<?php
/**
 * @version            5.0.0
 * @package            Joomla
 * @subpackage         Payment Form
 * @author             Tuan Pham Ngoc
 * @copyright          Copyright (C) 2010 - 2021 Ossolution Team
 * @license            GNU/GPL, see LICENSE.php
 */

defined('_JEXEC') or die;

use Joomla\CMS\Factory;
use Joomla\CMS\Language\Multilanguage;
use Joomla\CMS\Table\Table;
use Joomla\CMS\Uri\Uri;

class os_paypal extends POSFPayment
{
	/**
	 * Constructor functions, init some parameter
	 *
	 * @param   object  $params
	 */
	public function __construct($params, $config = [])
	{
		parent::__construct($params, $config);

		$this->mode = $params->get('paypal_mode');

		if ($this->mode)
		{
			$this->url = 'https://www.paypal.com/cgi-bin/webscr';
		}
		else
		{
			$this->url = 'https://www.sandbox.paypal.com/cgi-bin/webscr';
		}

		$locale = $params->get('paypal_locale');

		if (empty($locale))
		{
			if (Multilanguage::isEnabled())
			{
				$locale = Factory::getLanguage()->getTag();
				$locale = str_replace('-', '_', $locale);
			}
			else
			{
				$locale = 'en_US';
			}
		}

		$this->setParameter('business', $params->get('paypal_id'));
		$this->setParameter('rm', 2);
		$this->setParameter('cmd', '_xclick');
		$this->setParameter('no_shipping', 1);
		$this->setParameter('no_note', 1);
		$this->setParameter('lc', $locale);
		$this->setParameter('charset', 'utf-8');
		$this->setParameter('tax', 0);
	}

	/**
	 * Process Payment
	 *
	 * @param   object  $row
	 * @param   array   $data
	 */
	public function processPayment($row, $data)
	{
		$Itemid  = Factory::getApplication()->input->getInt('Itemid', 0);
		$siteUrl = Uri::base();

		$db    = Factory::getDbo();
		$query = $db->getQuery(true);
		$query->select('paypal_account')
			->from('#__pf_forms')
			->where('id = ' . $row->form_id);
		$db->setQuery($query);
		$payPalAccount = $db->loadResult();

		if ($payPalAccount)
		{
			$this->setParameter('business', $payPalAccount);
		}

		$this->setParameter('currency_code', $data['currency']);
		$this->setParameter('item_name', $data['item_name']);
		$this->setParameter('amount', round($data['amount'], 2));
		$this->setParameter('custom', $row->id);
		$this->setParameter('return', $siteUrl . 'index.php?option=com_pmform&view=complete&order_number=' . $row->order_number . '&Itemid=' . $Itemid);
		$this->setParameter('cancel_return', $siteUrl . 'index.php?option=com_pmform&view=cancel&id=' . $row->id . '&Itemid=' . $Itemid);
		$this->setParameter('notify_url', $siteUrl . 'index.php?option=com_pmform&task=payment_confirm&payment_method=os_paypal' . PMFormHelper::getLangLink());
		$this->setParameter('address1', $row->address);
		$this->setParameter('address2', $row->address2);
		$this->setParameter('city', $row->city);
		$this->setParameter('country', $data['country']);
		$this->setParameter('first_name', $row->first_name);
		$this->setParameter('last_name', $row->last_name);
		$this->setParameter('state', $row->state);
		$this->setParameter('zip', $row->zip);
		$this->setParameter('email', $row->email);


		$this->renderRedirectForm();
	}

	/**
	 * Verify payment
	 *
	 * @return bool
	 */
	public function verifyPayment()
	{
		$ret = $this->validate();

		if ($ret)
		{
			$id            = $this->notificationData['custom'];
			$transactionId = $this->notificationData['txn_id'];
			$amount        = $this->notificationData['mc_gross'];

			if ($amount < 0)
			{
				$this->logGatewayData('Invalid Amount:' . $amount);

				return false;
			}

			$row = Table::getInstance('pmform', 'Payment');

			if (!$row->load($id))
			{
				$this->logGatewayData('Invalid Record ID:' . $id);

				return false;
			}

			if ($row->published)
			{
				$this->logGatewayData('Invalid Status, already published before');

				return false;
			}

			// Validate PayPal ID
			$db    = Factory::getDbo();
			$query = $db->getQuery(true);
			$query->select('paypal_account')
				->from('#__pf_forms')
				->where('id = ' . $row->form_id);
			$db->setQuery($query);
			$paypalId = trim($db->loadResult());

			if (!$paypalId)
			{
				$paypalId = $this->params->get('paypal_id');
			}

			$receiverEmail = strtoupper($this->notificationData['receiver_email']);
			$receiverId    = strtoupper($this->notificationData['receiver_id']);
			$business      = strtoupper($this->notificationData['business']);
			$paypalId      = strtoupper($paypalId);

			if ($receiverEmail != $paypalId && $receiverId != $paypalId && $business != $paypalId)
			{
				$this->logGatewayData(sprintf('Invalid receiver. Expect %s, Get %s', $paypalId, $receiverEmail));

				return false;
			}

			// Validate currency
			$receivedPaymentCurrency = strtoupper($this->notificationData['mc_currency']);

			if ($receivedPaymentCurrency != strtoupper($row->payment_currency))
			{
				$this->logGatewayData(sprintf('Invalid Currency. Expect %s, Get %s', strtoupper($row->payment_currency), $receivedPaymentCurrency));

				return false;
			}

			// Validate payment amount
			$receivedAmount = floatval($this->notificationData['mc_gross']);

			// Accept 0.05$ difference
			if (($row->payment_amount - $receivedAmount) > 0.05)
			{
				$this->logGatewayData(sprintf('Invalid Payment Amount. Expect %s, Get %s', $row->payment_amount, $receivedAmount));

				return false;
			}

			$this->onPaymentSuccess($row, $transactionId);

			return true;
		}
	}

	/**
	 * Validate the post data from PayPal to our server
	 *
	 * @return string
	 */
	protected function validate()
	{
		JLoader::register('PaypalIPN', JPATH_ROOT . '/components/com_pmform/payments/paypal/PayPalIPN.php');

		$ipn = new PaypalIPN;

		// Use sandbox URL if test mode is configured
		if (!$this->mode)
		{
			$ipn->useSandbox();
		}

		// Disable use custom certs
		if ($this->params->get('use_local_certs', 0) == 0)
		{
			// Disable use custom certs
			$ipn->usePHPCerts();
		}

		$this->notificationData = $_POST;

		try
		{
			$valid = $ipn->verifyIPN();
			$this->logGatewayData($ipn->getResponse());

			if (!$this->mode || $valid)
			{
				return true;
			}

			return false;
		}
		catch (Exception $e)
		{
			$this->logGatewayData($e->getMessage());

			return false;
		}
	}
}